"""Shared utility functions."""

import contextlib
import functools
import glob
import inspect
import itertools
import os
import re
import subprocess
import sys
import threading
from collections.abc import (
    Callable,
    Iterable,
    MutableSequence,
)
from difflib import SequenceMatcher
from typing import (
    TYPE_CHECKING,
    Any,
    TextIO,
    TypeVar,
    Union,
    cast,
)

from rich.text import Text

from . import constants
from . import rich_utils as ru
from . import string_utils as su
from .types import (
    CmdOrSet,
    CmdOrSetT,
    UnboundChoicesProvider,
    UnboundCompleter,
)

if TYPE_CHECKING:  # pragma: no cover
    PopenTextIO = subprocess.Popen[str]
    from .argparse_utils import Cmd2ArgumentParser
else:
    PopenTextIO = subprocess.Popen

T = TypeVar("T")


def to_bool(val: Any) -> bool:
    """Convert anything to a boolean based on its value.

    Strings like "True", "true", "False", and "false" return True, True, False, and False
    respectively. All other values are converted using bool()

    :param val: value being converted
    :return: boolean value expressed in the passed in value
    :raises ValueError: if the string does not contain a value corresponding to a boolean value
    """
    if isinstance(val, str):
        if val.capitalize() == str(True):
            return True
        if val.capitalize() == str(False):
            return False
        raise ValueError("must be True or False (case-insensitive)")
    if isinstance(val, bool):
        return val
    return bool(val)


def optional_int(val: Any) -> int | None:
    """Convert a value to an integer or None if it's "None" (case-insensitive).

    :param val: value being converted
    :return: int or None
    :raises ValueError: if the value is not "None" and cannot be converted to an integer
    """
    if val is None:
        return None
    if isinstance(val, str) and val.lower() == "none":
        return None
    try:
        return int(val)
    except (ValueError, TypeError):
        raise ValueError("must be an integer or None (case-insensitive)") from None


class Settable:
    """Used to configure an attribute to be settable via the set command in the CLI."""

    def __init__(
        self,
        name: str,
        val_type: type[Any] | Callable[[Any], Any],
        description: str | Text,
        settable_object: object,
        *,
        settable_attrib_name: str | None = None,
        onchange_cb: Callable[[str, Any, Any], Any] | None = None,
        choices: Iterable[Any] | None = None,
        choices_provider: UnboundChoicesProvider[CmdOrSetT] | None = None,
        completer: UnboundCompleter[CmdOrSetT] | None = None,
    ) -> None:
        """Settable Initializer.

        :param name: The user-facing name for this setting in the CLI.
        :param val_type: A callable used to cast the string value from the CLI into its
                         proper type and validate it. This function should raise an
                         exception (like ValueError or TypeError) if the conversion or
                         validation fails, which will be caught and displayed to the user
                         by the set command. For example, setting this to int ensures the
                         input is a valid integer. Specifying bool automatically provides
                         completion for 'true' and 'false' and uses a built-in function
                         for conversion and validation.
        :param description: A concise string or rich Text object that describes the purpose of this setting.
        :param settable_object: The object that owns the attribute being made settable (e.g. self).
        :param settable_attrib_name: The name of the attribute on the settable_object that
                                     will be modified. This defaults to the value of the name
                                     parameter if not specified.
        :param onchange_cb: An optional function or method to call when the value of this
                            setting is altered by the set command. The callback is invoked
                            only if the new value is different from the old one.

                            It receives three arguments:
                                param_name: str - name of the parameter
                                old_value: Any - the parameter's old value
                                new_value: Any - the parameter's new value

        The following optional settings provide completion for a parameter's values.
        They correspond to the same settings in argparse-based completion. A maximum
        of one of these should be provided.

        :param choices: iterable of accepted values
        :param choices_provider: function that provides choices for this argument
        :param completer: completion function that provides choices for this argument
        """
        if val_type is bool:
            from .completion import Choices

            def get_bool_choices(_cmd2_self: CmdOrSet) -> Choices:
                """Tab complete lowercase boolean values."""
                return Choices.from_values(["true", "false"])

            val_type = to_bool
            choices_provider = get_bool_choices

        self.name = name
        self.val_type = val_type
        self.description = ru.rich_text_to_string(description) if isinstance(description, Text) else description
        self.settable_obj = settable_object
        self.settable_attrib_name = settable_attrib_name if settable_attrib_name is not None else name
        self.onchange_cb = onchange_cb
        self.choices = choices
        self.choices_provider = choices_provider
        self.completer = completer

    @property
    def value(self) -> Any:
        """Get the value of the settable attribute."""
        return getattr(self.settable_obj, self.settable_attrib_name)

    @value.setter
    def value(self, value: Any) -> None:
        """Set the settable attribute on the specified destination object.

        :param value: new value to set
        """
        # Run the value through its type function to handle any conversion or validation
        new_value = self.val_type(value)

        # Make sure new_value is a valid choice
        if self.choices is not None and new_value not in self.choices:
            choices_str = ", ".join(map(repr, self.choices))
            raise ValueError(f"invalid choice: {new_value!r} (choose from {choices_str})")

        # Try to update the settable's value
        orig_value = self.value
        setattr(self.settable_obj, self.settable_attrib_name, new_value)

        # Check if we need to call an onchange callback
        if orig_value != new_value and self.onchange_cb:
            self.onchange_cb(self.name, orig_value, new_value)


def is_text_file(file_path: str) -> bool:
    """Return if a file contains only ASCII or UTF-8 encoded text and isn't empty.

    :param file_path: path to the file being checked
    :return: True if the file is a non-empty text file, otherwise False
    :raises OSError: if file can't be read
    """
    expanded_path = os.path.abspath(os.path.expanduser(file_path.strip()))
    valid_text_file = False

    # Only need to check for utf-8 compliance since that covers ASCII, too
    try:
        with open(expanded_path, encoding="utf-8", errors="strict") as f:
            # Make sure the file has only utf-8 text and is not empty
            if sum(1 for _ in f) > 0:
                valid_text_file = True
    except OSError:
        raise
    except UnicodeDecodeError:
        # Not UTF-8
        pass

    return valid_text_file


def remove_duplicates(items: Iterable[T]) -> list[T]:
    """Remove duplicates from an iterable while preserving order of the items.

    :param items: the items being pruned of duplicates
    :return: a list containing only the unique items, in order
    """
    return list(dict.fromkeys(items))


def alphabetical_sort(items: Iterable[str]) -> list[str]:
    """Sorts an iterable of strings alphabetically.

    For example: ['a1', 'A11', 'A2', 'a22', 'a3']

    To sort a list in place, don't call this method, which makes a copy. Instead, do this:

    my_list.sort(key=norm_fold)

    :param items: the strings to sort
    :return: a sorted list
    """
    return sorted(items, key=su.norm_fold)


def try_int_or_force_to_lower_case(input_str: str) -> int | str:
    """Try to convert the passed-in string to an integer. If that fails, it converts it to lower case using norm_fold.

    :param input_str: string to convert
    :return: the string as an integer or a lower case version of the string.
    """
    try:
        return int(input_str)
    except ValueError:
        return su.norm_fold(input_str)


def natural_keys(input_str: str) -> list[int | str]:
    """Convert a string into a list of integers and strings to support natural sorting (see natural_sort).

    For example: natural_keys('abc123def') -> ['abc', '123', 'def']
    :param input_str: string to convert
    :return: list of strings and integers
    """
    return [try_int_or_force_to_lower_case(substr) for substr in re.split(r"(\d+)", input_str)]


def natural_sort(list_to_sort: Iterable[str]) -> list[str]:
    """Sorts a list of strings case insensitively as well as numerically.

    For example: ['a1', 'A2', 'a3', 'A11', 'a22']

    To sort a list in place, don't call this method, which makes a copy. Instead, do this:

    my_list.sort(key=natural_keys)

    :param list_to_sort: the list being sorted
    :return: the list sorted naturally
    """
    return sorted(list_to_sort, key=natural_keys)


def quote_specific_tokens(tokens: MutableSequence[str], tokens_to_quote: Iterable[str]) -> None:
    """Quote specific tokens in a list.

    :param tokens: token list being edited
    :param tokens_to_quote: the tokens, which if present in tokens, to quote
    """
    for i, token in enumerate(tokens):
        if token in tokens_to_quote:
            tokens[i] = su.quote(token)


def unquote_specific_tokens(tokens: MutableSequence[str], tokens_to_unquote: Iterable[str]) -> None:
    """Unquote specific tokens in a list.

    :param tokens: token list being edited
    :param tokens_to_unquote: the tokens, which if present in tokens, to unquote
    """
    for i, token in enumerate(tokens):
        unquoted_token = su.strip_quotes(token)
        if unquoted_token in tokens_to_unquote:
            tokens[i] = unquoted_token


def expand_user(token: str) -> str:
    """Wrap os.expanduser() to support expanding ~ in quoted strings.

    :param token: the string to expand
    """
    if token:
        if su.is_quoted(token):
            quote_char = token[0]
            token = su.strip_quotes(token)
        else:
            quote_char = ""

        token = os.path.expanduser(token)

        # Restore the quotes even if not needed to preserve what the user typed
        if quote_char:
            token = quote_char + token + quote_char

    return token


def expand_user_in_tokens(tokens: MutableSequence[str]) -> None:
    """Call expand_user() on all tokens in a list of strings.

    :param tokens: tokens to expand.
    """
    for index, _ in enumerate(tokens):
        tokens[index] = expand_user(tokens[index])


def find_editor() -> str | None:
    """Set cmd2.Cmd.DEFAULT_EDITOR. If EDITOR env variable is set, that will be used.

    Otherwise the function will look for a known editor in directories specified by PATH env variable.
    :return: Default editor or None.
    """
    editor = os.environ.get("EDITOR")
    if not editor:
        if sys.platform[:3] == "win":
            editors = ["edit", "code.cmd", "notepad++.exe", "notepad.exe"]
        else:
            editors = ["vim", "vi", "emacs", "nano", "pico", "joe", "code", "subl", "gedit", "kate"]

        # Get a list of every directory in the PATH environment variable and ignore symbolic links
        env_path = os.getenv("PATH")
        paths = [] if env_path is None else [p for p in env_path.split(os.path.pathsep) if not os.path.islink(p)]

        for possible_editor, path in itertools.product(editors, paths):
            editor_path = os.path.join(path, possible_editor)
            if os.path.isfile(editor_path) and os.access(editor_path, os.X_OK):
                if sys.platform[:3] == "win":
                    # Remove extension from Windows file names
                    editor = os.path.splitext(possible_editor)[0]
                else:
                    editor = possible_editor
                break
        else:
            editor = None

    return editor


def files_from_glob_pattern(pattern: str, access: int = os.F_OK) -> list[str]:
    """Return a list of file paths based on a glob pattern.

    Only files are returned, not directories, and optionally only files for which the user has a specified access to.

    :param pattern: file name or glob pattern
    :param access: file access type to verify (os.* where * is F_OK, R_OK, W_OK, or X_OK)
    :return: list of files matching the name or glob pattern
    """
    return [f for f in glob.glob(pattern) if os.path.isfile(f) and os.access(f, access)]


def files_from_glob_patterns(patterns: Iterable[str], access: int = os.F_OK) -> list[str]:
    """Return a list of file paths based on a list of glob patterns.

    Only files are returned, not directories, and optionally only files for which the user has a specified access to.

    :param patterns: Iterable of file names and/or glob patterns
    :param access: file access type to verify (os.* where * is F_OK, R_OK, W_OK, or X_OK)
    :return: list of files matching the names and/or glob patterns
    """
    files = []
    for pattern in patterns:
        matches = files_from_glob_pattern(pattern, access=access)
        files.extend(matches)
    return files


def get_exes_in_path(starts_with: str) -> list[str]:
    """Return names of executables in a user's path.

    :param starts_with: what the exes should start with. leave blank for all exes in path.
    :return: a list of matching exe names
    """
    # Purposely don't match any executable containing wildcards
    wildcards = ["*", "?"]
    for wildcard in wildcards:
        if wildcard in starts_with:
            return []

    # Get a list of every directory in the PATH environment variable and ignore symbolic links
    env_path = os.getenv("PATH")
    paths = [] if env_path is None else [p for p in env_path.split(os.path.pathsep) if not os.path.islink(p)]

    # Use a set to store exe names since there can be duplicates
    exes_set = set()

    # Find every executable file in the user's path that matches the pattern
    for path in paths:
        full_path = os.path.join(path, starts_with)
        matches = files_from_glob_pattern(full_path + "*", access=os.X_OK)

        for match in matches:
            exes_set.add(os.path.basename(match))

    return list(exes_set)


class StdSim:
    """Class to simulate behavior of sys.stdout or sys.stderr.

    Stores contents in internal buffer and optionally echos to the inner stream it is simulating.
    """

    def __init__(
        self,
        inner_stream: Union[TextIO, "StdSim"],
        *,
        echo: bool = False,
        encoding: str = "utf-8",
        errors: str = "replace",
    ) -> None:
        """StdSim Initializer.

        :param inner_stream: the wrapped stream. Should be a TextIO or StdSim instance.
        :param echo: if True, then all input will be echoed to inner_stream
        :param encoding: codec for encoding/decoding strings (defaults to utf-8)
        :param errors: how to handle encoding/decoding errors (defaults to replace)
        """
        self.inner_stream = inner_stream
        self.echo = echo
        self.encoding = encoding
        self.errors = errors
        self.pause_storage = False
        self.buffer = ByteBuf(self)

    def write(self, s: str) -> None:
        """Add str to internal bytes buffer and if echo is True, echo contents to inner stream.

        :param s: String to write to the stream
        """
        if not isinstance(s, str):
            raise TypeError(f"write() argument must be str, not {type(s)}")

        if not self.pause_storage:
            self.buffer.byte_buf += s.encode(encoding=self.encoding, errors=self.errors)
        if self.echo:
            self.inner_stream.write(s)

    def getvalue(self) -> str:
        """Get the internal contents as a str."""
        return self.buffer.byte_buf.decode(encoding=self.encoding, errors=self.errors)

    def getbytes(self) -> bytes:
        """Get the internal contents as bytes."""
        return bytes(self.buffer.byte_buf)

    def read(self, size: int | None = -1) -> str:
        """Read from the internal contents as a str and then clear them out.

        :param size: Number of bytes to read from the stream
        """
        if size is None or size == -1:
            result = self.getvalue()
            self.clear()
        else:
            result = self.buffer.byte_buf[:size].decode(encoding=self.encoding, errors=self.errors)
            self.buffer.byte_buf = self.buffer.byte_buf[size:]

        return result

    def readbytes(self) -> bytes:
        """Read from the internal contents as bytes and then clear them out."""
        result = self.getbytes()
        self.clear()
        return result

    def clear(self) -> None:
        """Clear the internal contents."""
        self.buffer.byte_buf.clear()

    def isatty(self) -> bool:
        """StdSim only considered an interactive stream if `echo` is True and `inner_stream` is a tty."""
        if self.echo:
            return self.inner_stream.isatty()
        return False

    @property
    def line_buffering(self) -> bool:
        """Handle when the inner stream doesn't have a line_buffering attribute.

        Which is the case when running unit tests because pytest sets stdout to a pytest EncodedFile object.
        """
        try:
            return bool(self.inner_stream.line_buffering)
        except AttributeError:
            return False

    def __getattr__(self, name: str) -> Any:
        """Forward attribute lookups to the inner stream for attributes not defined on this class."""
        return getattr(self.inner_stream, name)


class ByteBuf:
    """Used by StdSim to write binary data and stores the actual bytes written."""

    # Used to know when to flush the StdSim
    NEWLINES = (b"\n", b"\r")

    def __init__(self, std_sim_instance: StdSim) -> None:
        """Initialize the ByteBuf instance."""
        self.byte_buf = bytearray()
        self.std_sim_instance = std_sim_instance

    def write(self, b: bytes) -> None:
        """Add bytes to internal bytes buffer and if echo is True, echo contents to inner stream."""
        if not isinstance(b, bytes):
            raise TypeError(f"a bytes-like object is required, not {type(b)}")
        if not self.std_sim_instance.pause_storage:
            self.byte_buf += b
        if self.std_sim_instance.echo:
            self.std_sim_instance.inner_stream.buffer.write(b)

            # Since StdSim wraps TextIO streams, we will flush the stream if line buffering is on
            # and the bytes being written contain a new line character. This is helpful when StdSim
            # is being used to capture output of a shell command because it causes the output to print
            # to the screen more often than if we waited for the stream to flush its buffer.
            if self.std_sim_instance.line_buffering and any(newline in b for newline in ByteBuf.NEWLINES):
                self.std_sim_instance.flush()


class ProcReader:
    """Used to capture stdout and stderr from a Popen process if any of those were set to subprocess.PIPE.

    If neither are pipes, then the process will run normally and no output will be captured.
    """

    def __init__(self, proc: PopenTextIO, stdout: StdSim | TextIO, stderr: StdSim | TextIO) -> None:
        """ProcReader initializer.

        :param proc: the Popen process being read from
        :param stdout: the stream to write captured stdout
        :param stderr: the stream to write captured stderr.
        """
        self._proc = proc
        self._stdout = stdout
        self._stderr = stderr

        self._out_thread = threading.Thread(name="out_thread", target=self._reader_thread_func, kwargs={"read_stdout": True})

        self._err_thread = threading.Thread(name="err_thread", target=self._reader_thread_func, kwargs={"read_stdout": False})

        # Start the reader threads for pipes only
        if self._proc.stdout is not None:
            self._out_thread.start()
        if self._proc.stderr is not None:
            self._err_thread.start()

    def send_sigint(self) -> None:
        """Send a SIGINT to the process similar to if <Ctrl>+C were pressed."""
        import signal

        if sys.platform.startswith("win"):
            # cmd2 started the Windows process in a new process group. Therefore we must send
            # a CTRL_BREAK_EVENT since CTRL_C_EVENT signals cannot be generated for process groups.
            self._proc.send_signal(signal.CTRL_BREAK_EVENT)
        else:
            # Since cmd2 uses shell=True in its Popen calls, we need to send the SIGINT to
            # the whole process group to make sure it propagates further than the shell
            try:
                group_id = os.getpgid(self._proc.pid)
                os.killpg(group_id, signal.SIGINT)
            except ProcessLookupError:
                return

    def terminate(self) -> None:
        """Terminate the process."""
        self._proc.terminate()

    def wait(self) -> None:
        """Wait for the process to finish."""
        if self._out_thread.is_alive():
            self._out_thread.join()
        if self._err_thread.is_alive():
            self._err_thread.join()

        # Handle case where the process ended before the last read could be done.
        # This will return None for the streams that weren't pipes.
        out, err = self._proc.communicate()

        if out:
            self._write_bytes(self._stdout, out)
        if err:
            self._write_bytes(self._stderr, err)

    def _reader_thread_func(self, read_stdout: bool) -> None:
        """Thread function that reads a stream from the process.

        :param read_stdout: if True, then this thread deals with stdout. Otherwise it deals with stderr.
        """
        if read_stdout:
            read_stream = self._proc.stdout
            write_stream = self._stdout
        else:
            read_stream = self._proc.stderr
            write_stream = self._stderr

        # The thread should have been started only if this stream was a pipe
        if read_stream is None:
            raise ValueError("read_stream is None")

        # Run until process completes
        while self._proc.poll() is None:
            available = read_stream.peek()  # type: ignore[attr-defined]
            if available:
                read_stream.read(len(available))
                self._write_bytes(write_stream, available)

    @staticmethod
    def _write_bytes(stream: StdSim | TextIO, to_write: bytes | str) -> None:
        """Write bytes to a stream.

        :param stream: the stream being written to
        :param to_write: the bytes being written.
        """
        if isinstance(to_write, str):
            to_write = to_write.encode()

        # BrokenPipeError can occur if output is being piped to a process that closed
        with contextlib.suppress(BrokenPipeError):
            stream.buffer.write(to_write)


class ContextFlag:
    """A context manager which is also used as a boolean flag value within the default sigint handler.

    Its main use is as a flag to prevent the SIGINT handler in cmd2 from raising a KeyboardInterrupt
    while a critical code section has set the flag to True. Because signal handling is always done on the
    main thread, this class is not thread-safe since there is no need.
    """

    def __init__(self) -> None:
        """When this flag has a positive value, it is considered set. When it is 0, it is not set.

        It should never go below 0.
        """
        self._count = 0

    def __bool__(self) -> bool:
        """Define the truth value of an object when it is used in a boolean context."""
        return self._count > 0

    def __enter__(self) -> None:
        """When a with block is entered, the __enter__ method of the context manager is called."""
        self._count += 1

    def __exit__(self, *args: object) -> None:
        """When the execution flow exits a with statement block this is called, regardless of whether an exception occurred."""
        self._count -= 1
        if self._count < 0:
            raise ValueError("count has gone below 0")


class RedirectionSavedState:
    """Created by each command to store information required to restore state after redirection."""

    def __init__(
        self,
        self_stdout: "StdSim | TextIO",
        pipe_proc_reader: "ProcReader | None",
        saved_redirecting: bool,
    ) -> None:
        """RedirectionSavedState initializer.

        :param self_stdout: saved value of Cmd.stdout
        :param pipe_proc_reader: saved value of Cmd._cur_pipe_proc_reader
        :param saved_redirecting: saved value of Cmd._redirecting.
        """
        # Tells if command is redirecting
        self.redirecting = False

        # Used to restore Cmd.stdout after redirection ends
        self.saved_self_stdout = self_stdout

        # Used to restore values after command ends regardless of whether the command redirected
        self.saved_pipe_proc_reader = pipe_proc_reader
        self.saved_redirecting = saved_redirecting


def categorize(func: Callable[..., Any] | Iterable[Callable[..., Any]], category: str) -> None:
    """Categorize a function.

    The help command output will group the passed function under the
    specified category heading

    :param func: function or Iterable of functions to categorize
    :param category: category to put it in

    Example:
    ```py
    import cmd2
    class MyApp(cmd2.Cmd):
      def do_echo(self, arglist):
        self.poutput(' '.join(arglist)

      cmd2.utils.categorize(do_echo, "Text Processing")
    ```

    For an alternative approach to categorizing commands using a decorator, see [cmd2.decorators.with_category][]

    """
    funcs = func if isinstance(func, Iterable) else (func,)

    for cur_func in funcs:
        if inspect.ismethod(cur_func):
            setattr(cur_func.__func__, constants.COMMAND_ATTR_HELP_CATEGORY, category)
        else:
            setattr(cur_func, constants.COMMAND_ATTR_HELP_CATEGORY, category)


def get_defining_class(meth: Callable[..., Any]) -> type[Any] | None:
    """Attempt to resolve the class that defined a method.

    Inspired by implementation published here:
        https://stackoverflow.com/a/25959545/1956611

    :param meth: method to inspect
    :return: class type in which the supplied method was defined. None if it couldn't be resolved.
    """
    if isinstance(meth, functools.partial):
        return get_defining_class(meth.func)
    if inspect.ismethod(meth) or (
        inspect.isbuiltin(meth) and hasattr(meth, "__self__") and hasattr(meth.__self__, "__class__")
    ):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, "__func__", meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth), meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0])
        if isinstance(cls, type):
            return cls
    return cast(type, getattr(meth, "__objclass__", None))  # handle special descriptor objects


class CustomCompletionSettings:
    """Used by cmd2.Cmd.complete() to complete strings other than command arguments."""

    def __init__(self, parser: "Cmd2ArgumentParser", *, preserve_quotes: bool = False) -> None:
        """CustomCompletionSettings initializer.

        :param parser: arg parser defining format of string being completed
        :param preserve_quotes: if True, then quoted tokens will keep their quotes when processed by
                                ArgparseCompleter. This is helpful in cases when you're completing
                                flag-like tokens (e.g. -o, --option) and you don't want them to be
                                treated as argparse flags when quoted. Set this to True if you plan
                                on passing the string to argparse with the tokens still quoted.
        """
        self.parser = parser
        self.preserve_quotes = preserve_quotes


def strip_doc_annotations(doc: str) -> str:
    """Strip annotations from a docstring leaving only the text description.

    :param doc: documentation string
    """
    # Attempt to locate the first documentation block
    cmd_desc = ""
    found_first = False
    for doc_line in doc.splitlines():
        stripped_line = doc_line.strip()

        # Don't include :param type lines
        if stripped_line.startswith(":"):
            if found_first:
                break
        elif stripped_line:
            if found_first:
                cmd_desc += "\n"
            cmd_desc += stripped_line
            found_first = True
        elif found_first:
            break
    return cmd_desc


def similarity_function(s1: str, s2: str) -> float:
    """Ratio from s1,s2 may be different to s2,s1. We keep the max.

    See https://docs.python.org/3/library/difflib.html#difflib.SequenceMatcher.ratio
    """
    return max(SequenceMatcher(None, s1, s2).ratio(), SequenceMatcher(None, s2, s1).ratio())


MIN_SIMIL_TO_CONSIDER = 0.7


def suggest_similar(
    requested_command: str, options: Iterable[str], similarity_function_to_use: Callable[[str, str], float] | None = None
) -> str | None:
    """Given a requested command and an iterable of possible options returns the most similar (if any is similar).

    :param requested_command: The command entered by the user
    :param options: The list of available commands to search for the most similar
    :param similarity_function_to_use: An optional callable to use to compare commands
    :return: The most similar command or None if no one is similar
    """
    proposed_command = None
    best_simil = MIN_SIMIL_TO_CONSIDER
    requested_command_to_compare = requested_command.lower()
    similarity_function_to_use = similarity_function_to_use or similarity_function
    for each in options:
        simil = similarity_function_to_use(each.lower(), requested_command_to_compare)
        if best_simil < simil:
            best_simil = simil
            proposed_command = each
    return proposed_command


def get_types(func_or_method: Callable[..., Any]) -> tuple[dict[str, Any], Any]:
    """Use inspect.get_annotations() to extract type hints for parameters and return value.

    This is a thin convenience wrapper around inspect.get_annotations() that treats the return value
    annotation separately.

    :param func_or_method: Function or method to return the type hints for
    :return: tuple with first element being dictionary mapping param names to type hints
            and second element being the return type hint or None if there is no return value type hint
    :raises ValueError: if the `func_or_method` argument is not a valid object to pass to `inspect.get_annotations`
    """
    try:
        type_hints = inspect.get_annotations(func_or_method, eval_str=True)  # Get dictionary of type hints
    except TypeError as exc:
        raise ValueError("Argument passed to get_types should be a function or method") from exc
    ret_ann = type_hints.pop("return", None)  # Pop off the return annotation if it exists
    if inspect.ismethod(func_or_method):
        type_hints.pop("self", None)  # Pop off `self` hint for methods
    return type_hints, ret_ann


# Sorting keys for strings
ALPHABETICAL_SORT_KEY = su.norm_fold
NATURAL_SORT_KEY = natural_keys

# Application-wide sort key for strings
# Set it using cmd2.set_default_str_sort_key().
DEFAULT_STR_SORT_KEY: Callable[[str], str] = ALPHABETICAL_SORT_KEY


def set_default_str_sort_key(sort_key: Callable[[str], str]) -> None:
    """Set the application-wide sort key for strings."""
    global DEFAULT_STR_SORT_KEY  # noqa: PLW0603
    DEFAULT_STR_SORT_KEY = sort_key



"""Utilities for integrating prompt_toolkit with cmd2."""

import os
import re
from collections.abc import (
    Callable,
    Iterable,
)
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
)

from prompt_toolkit import print_formatted_text
from prompt_toolkit.application import get_app
from prompt_toolkit.completion import (
    Completer,
    Completion,
)
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.history import History
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.output.color_depth import ColorDepth
from rich.style import Style, StyleType

from . import (
    constants,
    utils,
)
from . import rich_utils as ru
from . import string_utils as su
from .styles import Cmd2Style

if TYPE_CHECKING:  # pragma: no cover
    from rich.color import Color

    from .cmd2 import Cmd


BASE_DELIMITERS = " \t\n" + "".join(constants.QUOTES) + "".join(constants.REDIRECTION_CHARS)

# prompt_toolkit accepts these standard ANSI color names directly
ANSI_NAMES = (
    "ansiblack",
    "ansired",
    "ansigreen",
    "ansiyellow",
    "ansiblue",
    "ansimagenta",
    "ansicyan",
    "ansiwhite",
    "ansibrightblack",
    "ansibrightred",
    "ansibrightgreen",
    "ansibrightyellow",
    "ansibrightblue",
    "ansibrightmagenta",
    "ansibrightcyan",
    "ansibrightwhite",
)


def pt_filter_style(text: str | ANSI) -> str | ANSI:
    """Strip styles if disallowed by ru.ALLOW_STYLE. Otherwise return an ANSI object.

    This function is intended specifically for text rendered by prompt-toolkit.
    """
    # We only use prompt-toolkit to write to a terminal. Therefore
    # we only have to check if ALLOW_STYLE is Never.
    if ru.ALLOW_STYLE == ru.AllowStyle.NEVER:
        raw_text = text.value if isinstance(text, ANSI) else text
        return su.strip_style(raw_text)

    # String must be an ANSI object for prompt-toolkit to render ANSI style sequences.
    return text if isinstance(text, ANSI) else ANSI(text)


def pt_resolve_color_depth() -> ColorDepth:
    """Determine the prompt-toolkit ColorDepth based on NO_COLOR and ru.ALLOW_STYLE."""
    if os.environ.get("NO_COLOR") or ru.ALLOW_STYLE == ru.AllowStyle.NEVER:
        return ColorDepth.MONOCHROME
    return ColorDepth.TRUE_COLOR


@lru_cache(maxsize=256)
def rich_to_pt_color(color: "Color | None") -> str:
    """Convert a rich Color object to a prompt_toolkit color string."""
    if not color or color.is_default:
        return "default"

    # Use prompt_toolkit's 16 standard ansi color names if applicable.
    # This prevents overriding terminal themes with absolute RGB values.
    if color.number is not None and 0 <= color.number <= 15:
        return ANSI_NAMES[color.number]

    # For 8-bit and truecolor, we fallback to hex RGB strings which prompt-toolkit supports natively
    c = color.get_truecolor()
    return f"#{c.red:02x}{c.green:02x}{c.blue:02x}"


@lru_cache(maxsize=1024)
def rich_to_pt_style(rich_style: StyleType) -> str:
    """Convert a rich Style object to a prompt_toolkit style string."""
    if not rich_style:
        return ""

    if isinstance(rich_style, str):
        rich_style = Style.parse(rich_style)

    parts = []

    fg_color = rich_to_pt_color(rich_style.color)
    parts.append(f"fg:{fg_color}")

    bg_color = rich_to_pt_color(rich_style.bgcolor)
    parts.append(f"bg:{bg_color}")

    if rich_style.bold is not None:
        parts.append("bold" if rich_style.bold else "nobold")
    if rich_style.underline is not None:
        parts.append("underline" if rich_style.underline else "nounderline")
    if rich_style.strike is not None:
        parts.append("strike" if rich_style.strike else "nostrike")
    if rich_style.italic is not None:
        parts.append("italic" if rich_style.italic else "noitalic")
    if rich_style.blink is not None:
        parts.append("blink" if rich_style.blink else "noblink")
    if rich_style.reverse is not None:
        parts.append("reverse" if rich_style.reverse else "noreverse")
    if rich_style.conceal is not None:
        parts.append("hidden" if rich_style.conceal else "nohidden")
    if rich_style.dim is not None:
        parts.append("dim" if rich_style.dim else "nodim")
    return " ".join(parts)


class Cmd2Completer(Completer):
    """Completer that delegates to cmd2's completion logic."""

    def __init__(
        self,
        cmd_app: "Cmd",
        custom_settings: utils.CustomCompletionSettings | None = None,
    ) -> None:
        """Initialize prompt_toolkit based completer class."""
        self._cmd_app = cmd_app
        self.custom_settings = custom_settings

    def get_completions(self, document: Document, complete_event: object) -> Iterable[Completion]:  # noqa: ARG002
        """Get completions for the current input."""
        # Find the beginning of the current word based on delimiters
        line = document.text
        cursor_pos = document.cursor_position

        # Define delimiters for completion to match cmd2/readline behavior
        delimiters = BASE_DELIMITERS
        delimiters += "".join(self._cmd_app.statement_parser.terminators)

        # Find last delimiter before cursor to determine the word being completed
        begidx = 0
        for i in range(cursor_pos - 1, -1, -1):
            if line[i] in delimiters:
                begidx = i + 1
                break

        endidx = cursor_pos
        text = line[begidx:endidx]

        completions = self._cmd_app.complete(
            text, line=line, begidx=begidx, endidx=endidx, custom_settings=self.custom_settings
        )

        if completions.error:
            print_formatted_text(pt_filter_style(completions.error))
            return

        # Print completion table if present
        if completions.table is not None:
            console = ru.Cmd2GeneralConsole(file=self._cmd_app.stdout)
            with console.capture() as capture:
                console.print(completions.table, end="", soft_wrap=False)
            print_formatted_text(pt_filter_style("\n" + capture.get()))

        if not completions:
            # Print hint if present
            if completions.hint:
                print_formatted_text(pt_filter_style(completions.hint))
            return

        # The length of the user's input minus any shortcut.
        search_text_length = len(text) - completions._search_text_offset

        # If matches require quoting but the word isn't quoted yet, we insert the
        # opening quote directly into the buffer. We do this because if any completions
        # change text before the cursor (like prepending a quote), prompt-toolkit will
        # not return a common prefix to the command line. By modifying the buffer
        # and returning early, we trigger a new completion cycle where the quote
        # is already present, allowing for proper common prefix calculation.
        if completions._add_opening_quote and search_text_length > 0:
            buffer = get_app().current_buffer

            buffer.cursor_left(search_text_length)
            buffer.insert_text(completions._quote_char)
            buffer.cursor_right(search_text_length)
            return

        # Return the completions
        for item in completions:
            # Set offset to the start of the current word to overwrite it with the completion
            start_position = -len(text)
            match_text = item.text

            # If we need a quote but didn't interrupt (because text was empty),
            # prepend the quote here so it's included in the insertion.
            if completions._add_opening_quote:
                match_text = (
                    match_text[: completions._search_text_offset]
                    + completions._quote_char
                    + match_text[completions._search_text_offset :]
                )

            # Finalize if there's only one match
            if len(completions) == 1 and completions.allow_finalization:
                # Close any open quote
                if completions._quote_char:
                    match_text += completions._quote_char

                # Add trailing space if the cursor is at the end of the line
                if endidx == len(line):
                    match_text += " "

            yield Completion(
                match_text,
                start_position=start_position,
                display=pt_filter_style(item.display),
                display_meta=pt_filter_style(item.display_meta),
            )


class Cmd2History(History):
    """A non-persistent, in-memory history buffer for prompt-toolkit.

    This class serves as the backing store for UI history navigation (e.g., arrowing
    through previous commands). It explicitly avoids handling persistence,
    deferring all permanent storage logic to the cmd2 application.
    """

    def __init__(self, history_strings: Iterable[str] | None = None) -> None:
        """Initialize the instance."""
        super().__init__()

        if history_strings:
            for string in history_strings:
                self.append_string(string)

        # Mark that self._loaded_strings is populated.
        self._loaded = True

    def append_string(self, string: str) -> None:
        """Override to filter our consecutive duplicates."""
        # History is sorted newest to oldest, so we compare to the first element.
        if string and (not self._loaded_strings or self._loaded_strings[0] != string):
            super().append_string(string)

    def store_string(self, string: str) -> None:
        """No-op: Persistent history data is stored in cmd2.Cmd.history."""

    def load_history_strings(self) -> Iterable[str]:
        """Yield strings from newest to oldest."""
        yield from self._loaded_strings

    def clear(self) -> None:
        """Clear the UI history navigation data."""
        self._loaded_strings.clear()


class Cmd2Lexer(Lexer):
    """Lexer that highlights cmd2 command names, aliases, and macros."""

    # Use the 'class:' prefix to look up styles in the prompt-toolkit theme
    COMMAND_STYLE = f"class:{Cmd2Style.LEXER_COMMAND}"
    ALIAS_STYLE = f"class:{Cmd2Style.LEXER_ALIAS}"
    MACRO_STYLE = f"class:{Cmd2Style.LEXER_MACRO}"
    FLAG_STYLE = f"class:{Cmd2Style.LEXER_FLAG}"
    ARGUMENT_STYLE = f"class:{Cmd2Style.LEXER_ARGUMENT}"

    def __init__(
        self,
        cmd_app: "Cmd",
    ) -> None:
        """Initialize the Lexer.

        :param cmd_app: cmd2.Cmd instance
        """
        super().__init__()
        self._cmd_app = cmd_app

    def lex_document(self, document: Document) -> Callable[[int], Any]:
        """Lex the document."""
        # Get redirection tokens and terminators to avoid highlighting them as values
        exclude_tokens = set(constants.REDIRECTION_TOKENS)
        exclude_tokens.update(self._cmd_app.statement_parser.terminators)
        arg_pattern = re.compile(r'(\s+)|(--?[^\s\'"]+)|("[^"]*"?|\'[^\']*\'?)|([^\s\'"]+)')

        def highlight_args(text: str, tokens: list[tuple[str, str]]) -> None:
            """Highlight arguments in a string."""
            for m in arg_pattern.finditer(text):
                space, flag, quoted, word = m.groups()
                match_text = m.group(0)

                if space:
                    tokens.append(("", match_text))
                elif flag:
                    tokens.append((self.FLAG_STYLE, match_text))
                elif (quoted or word) and match_text not in exclude_tokens:
                    tokens.append((self.ARGUMENT_STYLE, match_text))
                else:
                    tokens.append(("", match_text))

        def get_line(lineno: int) -> list[tuple[str, str]]:
            """Return the tokens for the given line number."""
            line = document.lines[lineno]
            tokens: list[tuple[str, str]] = []

            # No syntax highlighting if styles are disallowed
            if ru.ALLOW_STYLE == ru.AllowStyle.NEVER:
                tokens.append(("", line))
                return tokens

            # Only attempt to match a command on the first line
            if lineno == 0:
                # Use cmd2's command pattern to find the first word (the command)
                match = self._cmd_app.statement_parser._command_pattern.search(line)
                if match:
                    # Group 1 is the command, Group 2 is the character(s) that terminated the command match
                    command = match.group(1)
                    cmd_start = match.start(1)
                    cmd_end = match.end(1)

                    # Add any leading whitespace
                    if cmd_start > 0:
                        tokens.append(("", line[:cmd_start]))

                    if command:
                        # Determine the style for the command
                        shortcut_found = False
                        for shortcut, _ in self._cmd_app.statement_parser.shortcuts:
                            if command.startswith(shortcut):
                                # Add the shortcut with the command style
                                tokens.append((self.COMMAND_STYLE, shortcut))

                                # If there's more in the command word, it's an argument
                                if len(command) > len(shortcut):
                                    tokens.append((self.ARGUMENT_STYLE, command[len(shortcut) :]))

                                shortcut_found = True
                                break

                        if not shortcut_found:
                            style = ""
                            if command in self._cmd_app.get_all_commands():
                                style = self.COMMAND_STYLE
                            elif command in self._cmd_app.aliases:
                                style = self.ALIAS_STYLE
                            elif command in self._cmd_app.macros:
                                style = self.MACRO_STYLE

                            # Add the command with the determined style
                            tokens.append((style, command))

                    # Add the rest of the line as arguments
                    if cmd_end < len(line):
                        highlight_args(line[cmd_end:], tokens)
                else:
                    # No command match found on the first line
                    tokens.append(("", line))
            else:
                # All other lines are treated as arguments
                highlight_args(line, tokens)

            return tokens

        return get_line



"""Defines custom Rich styles and their corresponding names for cmd2.

This module provides a centralized and discoverable way to manage Rich styles
used within the cmd2 framework. It defines a StrEnum for style names and a
dictionary that maps these names to their default style objects.

**Notes**

Cmd2 uses Rich for its terminal output, and while this module defines a set of
cmd2-specific styles, it's important to understand that these aren't the only
styles that can appear. Components like Rich tracebacks and the rich-argparse
library, which cmd2 uses for its help output, also apply their own built-in
styles. Additionally, app developers may use other Rich objects that have
their own default styles.

For a complete theming experience, you can create a custom theme that includes
styles from Rich and rich-argparse. The `cmd2.theme.update_theme()` function
automatically updates rich-argparse's styles with any custom styles provided in
your theme dictionary, so you don't have to modify them directly.

You can find Rich's default styles in the `rich.default_styles` module.
For rich-argparse, the style names are defined in the
`rich_argparse.RichHelpFormatter.styles` dictionary.

For prompt-toolkit default styles, see:
https://github.com/prompt-toolkit/python-prompt-toolkit/blob/main/src/prompt_toolkit/styles/defaults.py
"""

from enum import StrEnum

from rich.style import (
    Style,
    StyleType,
)
from rich_argparse import RichHelpFormatter

from .colors import Color


class Cmd2Style(StrEnum):
    """An enumeration of the names of custom Rich styles used in cmd2.

    Using this enum allows for autocompletion and prevents typos when
    referencing cmd2-specific styles.

    This StrEnum is tightly coupled with `DEFAULT_CMD2_STYLES`. Any name
    added here must have a corresponding style definition there.
    """

    COMMAND_LINE = "cmd2.example"  # Command line examples in help text
    COMPLETION_MENU = "cmd2.completion-menu"  # Base style for the entire completion menu container (sets the background)
    COMPLETION_MENU_COMPLETION = "cmd2.completion-menu.completion"  # Style for an individual, non-selected completion item
    COMPLETION_MENU_CURRENT = "cmd2.completion-menu.completion.current"  # Style for the currently selected completion item
    COMPLETION_MENU_META = "cmd2.completion-menu.meta.completion"  # Style for meta information shown alongside a completion
    COMPLETION_MENU_META_CURRENT = "cmd2.completion-menu.meta.completion.current"  # Style for meta information of current item
    ERROR = "cmd2.error"  # Error text (used by perror())
    HELP_HEADER = "cmd2.help.header"  # Help table header text
    HELP_LEADER = "cmd2.help.leader"  # Text right before the help tables are listed
    LEXER_COMMAND = "cmd2.lexer.command"  # Lexer color for commands
    LEXER_ALIAS = "cmd2.lexer.alias"  # Lexer color for aliases
    LEXER_MACRO = "cmd2.lexer.macro"  # Lexer color for macros
    LEXER_FLAG = "cmd2.lexer.flag"  # Lexer color for flags
    LEXER_ARGUMENT = "cmd2.lexer.argument"  # Lexer color for arguments
    SUCCESS = "cmd2.success"  # Success text (used by psuccess())
    TABLE_BORDER = "cmd2.table_border"  # Applied to cmd2's table borders
    WARNING = "cmd2.warning"  # Warning text (used by pwarning())


# Default styles used by cmd2. Used to perform theme resets.
# Tightly coupled with the Cmd2Style enum.
DEFAULT_CMD2_STYLES: dict[str, StyleType] = {
    Cmd2Style.COMMAND_LINE: Style(color=Color.CYAN, bold=True),
    Cmd2Style.COMPLETION_MENU: Style(color="#000000", bgcolor="#bbbbbb"),  # prompt-toolkit default
    Cmd2Style.COMPLETION_MENU_COMPLETION: Style(),  # prompt-toolkit default
    # Defined with swapped FG/BG colors because prompt-toolkit applies 'reverse' by default.
    # This yields a black-on-green highlight in color mode, while maintaining a fallback
    # reverse-video highlight when colors are disabled (e.g. NO_COLOR=1 or allow_style=never).
    Cmd2Style.COMPLETION_MENU_CURRENT: Style(color=Color.GREEN, bgcolor=Color.BLACK),
    Cmd2Style.COMPLETION_MENU_META: Style(color="#000000", bgcolor="#bbbbbb"),  # prompt-toolkit default
    Cmd2Style.COMPLETION_MENU_META_CURRENT: Style(color=Color.BLACK, bgcolor=Color.BRIGHT_GREEN),
    Cmd2Style.ERROR: Style(color=Color.BRIGHT_RED),
    Cmd2Style.HELP_HEADER: Style(color=Color.BRIGHT_GREEN),
    Cmd2Style.HELP_LEADER: Style(color=Color.CYAN),
    Cmd2Style.LEXER_COMMAND: Style(color=Color.GREEN),
    Cmd2Style.LEXER_ALIAS: Style(color=Color.CYAN),
    Cmd2Style.LEXER_MACRO: Style(color=Color.MAGENTA),
    Cmd2Style.LEXER_FLAG: Style(color=Color.RED),
    Cmd2Style.LEXER_ARGUMENT: Style(color=Color.YELLOW),
    Cmd2Style.SUCCESS: Style(color=Color.GREEN),
    Cmd2Style.TABLE_BORDER: Style(color=Color.BRIGHT_GREEN),
    Cmd2Style.WARNING: Style(color=Color.BRIGHT_YELLOW),
}

# Default styles for argparse output. Used to perform theme resets.
# Any cmd2-specific settings or overrides should be added to this dictionary.
DEFAULT_ARGPARSE_STYLES = RichHelpFormatter.styles.copy()



"""Provides a centralized theming system for cmd2.

This module manages the global theme used for both Rich terminal output and
prompt-toolkit interactive components. It ensures that styling is consistent
and synchronized across the entire application when the theme is updated using
the following strategy.

1. Rich consoles use a persistent Theme object that is updated in-place.
2. prompt-toolkit integration in cmd2 uses a DynamicStyle wrapper around a
   callable that returns the current prompt-toolkit theme.

To prevent Rich's built-in styles (like 'bold', or 'italic') from
polluting the prompt-toolkit namespace or colliding with its own internal
names, only styles starting with registered prefixes (e.g., 'cmd2.') or
those explicitly mapped to UI elements are synchronized to the
prompt-toolkit theme.
"""

from collections.abc import (
    Iterable,
    Mapping,
)
from typing import cast

from prompt_toolkit.styles import Style as PtStyle
from rich.style import (
    Style,
    StyleType,
)
from rich.theme import Theme

from .pt_utils import rich_to_pt_style
from .rich_utils import Cmd2HelpFormatter
from .styles import (
    DEFAULT_ARGPARSE_STYLES,
    DEFAULT_CMD2_STYLES,
    Cmd2Style,
)

# The application-wide theme, defined using Rich's styling system.
# Use get_theme() to access it.
# Use reset_theme() and update_theme() to modify it.
_THEME: Theme | None = None

# The prompt-toolkit version of the theme, synchronized from the Rich theme.
# Use get_pt_theme() to access it.
_PT_THEME: PtStyle | None = None

# Maps style names to internal UI component names used by prompt-toolkit.
# This allows developers to use application-specific style names in update_theme()
# while ensuring the underlying prompt-toolkit UI is styled correctly.
# Use register_pt_mapping() and unregister_pt_mapping() to manage these mappings.
#
# Presence in this mapping, even with an empty set of UI component names, flags the
# style for synchronization to the prompt-toolkit theme. Use register_synchronized_style()
# and unregister_synchronized_style() to manage synchronization for styles that lack a
# registered prefix and do not require specific UI component mappings.
_PT_UI_MAP: dict[str, set[str]] = {
    Cmd2Style.COMPLETION_MENU: {"completion-menu"},
    Cmd2Style.COMPLETION_MENU_COMPLETION: {"completion-menu.completion"},
    Cmd2Style.COMPLETION_MENU_CURRENT: {"completion-menu.completion.current"},
    Cmd2Style.COMPLETION_MENU_META: {"completion-menu.meta.completion"},
    Cmd2Style.COMPLETION_MENU_META_CURRENT: {
        "completion-menu.meta.completion.current",
        "completion-menu.multi-column-meta",
    },
}

# Rich styles that start with one of these prefixes are automatically
# synchronized to the prompt-toolkit theme. Use register_synchronized_prefix()
# and unregister_synchronized_prefix() to modify this set.
_SYNCHRONIZED_PREFIXES: set[str] = {"cmd2."}


def get_theme() -> Theme:
    """Get the application-wide Rich theme. Initializes it on the first call."""
    if _THEME is None:
        reset_theme()
    return cast(Theme, _THEME)


def get_pt_theme() -> PtStyle:
    """Get the application-wide prompt-toolkit style. Initializes it on the first call."""
    if _PT_THEME is None:
        reset_theme()
    return cast(PtStyle, _PT_THEME)


def reset_theme() -> None:
    """Reset the application-wide theme to its initial state.

    This function performs an in-place reset of the existing Rich theme's
    styles. This ensures that any Console objects already using the theme
    will reflect the changes immediately without needing to be recreated.

    Changes are automatically propagated to all synchronized components.
    """
    global _THEME  # noqa: PLW0603

    # Include default styles from cmd2, rich-argparse, and Rich.
    styles = DEFAULT_CMD2_STYLES.copy()
    styles.update(DEFAULT_ARGPARSE_STYLES)
    default_theme = Theme(styles, inherit=True)

    if _THEME is None:
        # Initial assignment
        _THEME = default_theme
    else:
        # Perform in-place reset to preserve existing references
        _THEME.styles.clear()
        _THEME.styles.update(default_theme.styles)

    _sync_all()


def update_theme(styles: Mapping[str, StyleType]) -> None:
    """Update the existing theme.

    This function performs an in-place update of the existing Rich theme's
    styles. This ensures that any Console objects already using the theme
    will reflect the changes immediately without needing to be recreated.

    Changes are automatically propagated to all synchronized components.

    :param styles: mapping of style names to styles
    """
    # Convert any string styles to Style objects
    parsed_styles = {name: style if isinstance(style, Style) else Style.parse(style) for name, style in styles.items()}

    # Perform in-place update to preserve existing references
    get_theme().styles.update(parsed_styles)

    _sync_all()


def _sync_all() -> None:
    """Propagate the global theme to rich-argparse and prompt-toolkit.

    If the theme hasn't been initialized yet, this is a no-op.
    """
    if _THEME is None:
        return

    # Synchronize rich-argparse styles
    for name in Cmd2HelpFormatter.styles.keys() & _THEME.styles.keys():
        Cmd2HelpFormatter.styles[name] = _THEME.styles[name]

    # Synchronize the prompt-toolkit theme
    _sync_pt_theme()


def _sync_pt_theme() -> None:
    """Build a new global prompt-toolkit style object based on the current Rich theme.

    If the theme hasn't been initialized yet, this is a no-op.
    """
    if _THEME is None:
        return

    style_rules: list[tuple[str, str]] = []

    for name, rich_style in _THEME.styles.items():
        # Only synchronize if it has a registered prefix or mapped UI component.
        is_framework_style = any(name.startswith(p) for p in _SYNCHRONIZED_PREFIXES)
        is_mapped_style = name in _PT_UI_MAP

        if is_framework_style or is_mapped_style:
            pt_style_str = rich_to_pt_style(rich_style)

            # Register the style name as a prompt-toolkit class (accessible via 'class:name')
            style_rules.append((name, pt_style_str))

            # Add any prompt-toolkit UI component names from the map (e.g., 'completion-menu')
            if is_mapped_style:
                style_rules.extend((pt_name, pt_style_str) for pt_name in _PT_UI_MAP[name])

    global _PT_THEME  # noqa: PLW0603
    _PT_THEME = PtStyle(style_rules)


def register_pt_mapping(style_name: str, pt_ui_names: str | Iterable[str]) -> None:
    """Map a Rich theme style name to one or more prompt-toolkit UI components.

    This enables styling of prompt-toolkit's internal elements (such as the
    completion menu) using styles in the application's Rich theme.

    Registering a mapping also flags the style for synchronization to the
    prompt-toolkit theme, making it accessible via 'class:style_name'.

    :param style_name: The style name used in the Rich theme.
    :param pt_ui_names: One or more prompt-toolkit UI component names (e.g., 'completion-menu').
    """
    if isinstance(pt_ui_names, str):
        pt_ui_names = [pt_ui_names]

    # Register the style in the map.
    if style_name not in _PT_UI_MAP:
        _PT_UI_MAP[style_name] = set()
        changed = True
    else:
        changed = False

    # Add UI mappings, excluding 'style_name' which the sync handles by default.
    original_size = len(_PT_UI_MAP[style_name])
    _PT_UI_MAP[style_name].update(n for n in pt_ui_names if n != style_name)

    if len(_PT_UI_MAP[style_name]) != original_size:
        changed = True

    # Trigger a re-sync if the theme is already initialized
    if changed and _PT_THEME is not None:
        _sync_pt_theme()


def unregister_pt_mapping(style_name: str, pt_ui_names: str | Iterable[str]) -> None:
    """Remove one or more prompt-toolkit UI component mappings.

    The style itself remains in the synchronization mapping (even if no
    UI component mappings remain), ensuring it continues to be synchronized
    to the prompt-toolkit theme.

    To completely remove a style from synchronization, use
    unregister_synchronized_style().

    :param style_name: The style name used in the Rich theme.
    :param pt_ui_names: Specific UI component(s) to unmap.
    """
    if style_name not in _PT_UI_MAP:
        return

    if isinstance(pt_ui_names, str):
        pt_ui_names = [pt_ui_names]

    original_size = len(_PT_UI_MAP[style_name])
    for name in pt_ui_names:
        _PT_UI_MAP[style_name].discard(name)

    changed = len(_PT_UI_MAP[style_name]) != original_size

    # Trigger a re-sync if the theme is already initialized
    if changed and _PT_THEME is not None:
        _sync_pt_theme()


def register_synchronized_style(style_name: str) -> None:
    """Register a Rich theme style for synchronization with prompt-toolkit.

    This ensures that the style is synchronized to the prompt-toolkit theme
    (accessible via 'class:style_name') even if it does not begin with a
    registered prefix.

    :param style_name: The style name used in the Rich theme.
    """
    register_pt_mapping(style_name, [])


def unregister_synchronized_style(style_name: str) -> None:
    """Stop synchronizing a Rich theme style with prompt-toolkit.

    This removes the style and all its mappings entirely from the
    prompt-toolkit theme synchronization.

    :param style_name: The style name to unregister.
    """
    if style_name in _PT_UI_MAP:
        del _PT_UI_MAP[style_name]

        # Trigger a re-sync if the theme is already initialized
        if _PT_THEME is not None:
            _sync_pt_theme()


def register_synchronized_prefix(prefix: str) -> None:
    """Register a prefix whose styles will be synchronized to the prompt-toolkit theme.

    The prefix must include any desired delimiters (e.g., 'myapp.' or 'plugin-').

    :param prefix: The prefix string. Must be at least 1 character.
    :raises ValueError: If the prefix is empty.
    """
    if not prefix:
        raise ValueError("Prefix cannot be empty.")

    if prefix not in _SYNCHRONIZED_PREFIXES:
        _SYNCHRONIZED_PREFIXES.add(prefix)

        # Trigger a re-sync if the theme is already initialized
        if _PT_THEME is not None:
            _sync_pt_theme()


def unregister_synchronized_prefix(prefix: str) -> None:
    """Stop synchronizing styles starting with the given prefix.

    :param prefix: The prefix string to remove.
    """
    if prefix in _SYNCHRONIZED_PREFIXES:
        _SYNCHRONIZED_PREFIXES.remove(prefix)

        # Trigger a re-sync if the theme is already initialized
        if _PT_THEME is not None:
            _sync_pt_theme()



"""Provides classes and functions related to command-line completion."""

import copy
import re
from collections.abc import (
    Iterable,
    Iterator,
    Sequence,
)
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Any,
    Self,
    cast,
    overload,
)

from rich.protocol import is_renderable
from rich.table import Table

from . import rich_utils as ru
from . import string_utils as su


class _UnsetStr(str):
    """Internal sentinel to distinguish between an unset and an explicit empty string."""

    __slots__ = ()


_UNSET_STR = _UnsetStr("")


@dataclass(frozen=True, slots=True, kw_only=True)
class CompletionItem:
    """A single completion result."""

    # Regular expression to identify whitespace characters that are rendered as
    # control sequences (like ^J or ^I) in the completion menu.
    _CONTROL_WHITESPACE_RE = re.compile(r"\r\n|[\n\r\t\f\v]")

    # The source input for the completion. This is used to initialize the 'text'
    # field (defaults to str(value)). The original object is also preserved to
    # support object-based validation when this CompletionItem is used as an
    # argparse choice.
    value: Any = field(kw_only=False)

    # The string matched against user input and inserted into the command line.
    # Defaults to str(value). This should only be set manually if this
    # CompletionItem is used as an argparse choice and you want the choice
    # string to differ from str(value).
    text: str = _UNSET_STR

    # Optional string for displaying the completion differently in the completion menu.
    # This can contain ANSI style sequences. A plain version is stored in display_plain.
    # If not provided, defaults to the (possibly computed) value of 'text'.
    display: str = _UNSET_STR

    # Optional meta information about completion which displays in the completion menu.
    # This can contain ANSI style sequences. A plain version is stored in display_meta_plain.
    display_meta: str = ""

    # Optional data for completion tables. Length must match the associated argparse
    # argument's table_columns. This is stored internally as a tuple.
    table_data: Sequence[Any] = field(default_factory=tuple)

    # Plain text versions of display fields (stripped of ANSI) for sorting/filtering.
    display_plain: str = field(default="", init=False)
    display_meta_plain: str = field(default="", init=False)

    @classmethod
    def _clean_display(cls, val: str) -> str:
        """Clean a string for display in the completion menu.

        This replaces whitespace characters that are rendered as
        control sequences (like ^J or ^I) with spaces.

        :param val: string to be cleaned
        :return: the cleaned string
        """
        return cls._CONTROL_WHITESPACE_RE.sub(" ", val)

    def __post_init__(self) -> None:
        """Finalize the object after initialization.

        By using the sentinel pattern to distinguish between a field that was never
        set and one explicitly blanked out, this handles the two-stage lifecycle:

        1. Initial creation (usually by a developer-provided choices_provider or completer).
        2. Post-processing by cmd2 via dataclasses.replace(), which may modify fields or
           explicitly set them to empty strings.
        """
        # If the completion string was not provided, derive it from value.
        if isinstance(self.text, _UnsetStr):
            object.__setattr__(self, "text", str(self.value))

        # If the display string was not provided, use text.
        if isinstance(self.display, _UnsetStr):
            object.__setattr__(self, "display", self.text)

        # Clean display and display_meta
        object.__setattr__(self, "display", self._clean_display(self.display))
        object.__setattr__(self, "display_meta", self._clean_display(self.display_meta))

        # Create plain text versions by stripping ANSI sequences.
        # These are stored as attributes for fast access during sorting/filtering.
        object.__setattr__(self, "display_plain", su.strip_style(self.display))
        object.__setattr__(self, "display_meta_plain", su.strip_style(self.display_meta))

        # Make sure all table data objects are renderable by a Rich table.
        renderable_data = [obj if is_renderable(obj) else str(obj) for obj in self.table_data]

        # Convert strings containing ANSI style sequences to Rich Text objects for correct display width.
        object.__setattr__(
            self,
            "table_data",
            ru.prepare_objects_for_rendering(*renderable_data),
        )

    def __deepcopy__(self, memo: dict[int, Any]) -> "CompletionItem":
        """Return a shallow copy of this CompletionItem during a deepcopy operation.

        This is necessary because cmd2 deepcopies argument parsers to keep them unique
        across command instances. This override prevents the deepcopying of
        CompletionItems stored within a parser's 'choices' list.

        Since the 'value' and 'table_data' fields may contain complex objects which
        should not be deep copied, a shallow copy ensures the original object
        references are preserved.
        """
        return copy.copy(self)

    def __str__(self) -> str:
        """Return the completion text."""
        return self.text

    def __eq__(self, other: object) -> bool:
        """Compare this CompletionItem for equality.

        Identity is determined by value, text, display, and display_meta.
        table_data is excluded from equality checks to ensure that items
        with the same functional value are treated as duplicates.

        Also supports comparison against non-CompletionItems to facilitate argparse
        choices validation.
        """
        if isinstance(other, CompletionItem):
            return (
                self.value == other.value
                and self.text == other.text
                and self.display == other.display
                and self.display_meta == other.display_meta
            )

        # This supports argparse validation when a CompletionItem is used as a choice
        return bool(self.value == other)

    def __hash__(self) -> int:
        """Return a hash of the item's identity fields."""
        return hash((self.value, self.text, self.display, self.display_meta))


@dataclass(frozen=True, slots=True, kw_only=True)
class CompletionResultsBase:
    """Base class for results containing a collection of CompletionItems."""

    # Regular expression to identify strings that we should sort numerically
    _NUMERIC_RE = re.compile(
        r"""
        ^              # Start of string
        [-+]?          # Optional sign
        (?:            # Start of non-capturing group
            \d+\.?\d*  # Matches 123 or 123. or 123.45
            |          # OR
            \.\d+      # Matches .45
        )              # End of group
        $              # End of string
    """,
        re.VERBOSE,
    )

    # The collection of CompletionItems. This is stored internally as a tuple.
    items: Sequence[CompletionItem] = field(default_factory=tuple, kw_only=False)

    # If True, indicates the items are already provided in the desired display order.
    # If False, items will be sorted by their display value during initialization.
    is_sorted: bool = False

    # True if every item in this collection has a numeric display string.
    # Used for sorting and alignment.
    numeric_display: bool = field(default=False, init=False)

    def __post_init__(self) -> None:
        """Finalize the object after initialization."""
        from . import utils

        unique_items = utils.remove_duplicates(self.items)

        # Determine if all items have numeric display strings
        numeric_display = bool(unique_items) and all(self._NUMERIC_RE.match(i.display_plain) for i in unique_items)
        object.__setattr__(self, "numeric_display", numeric_display)

        if not self.is_sorted:
            if self.numeric_display:
                # Sort numerically
                unique_items.sort(key=lambda item: float(item.display_plain))
            else:
                # Standard string sort
                unique_items.sort(key=lambda item: utils.DEFAULT_STR_SORT_KEY(item.display_plain))

            object.__setattr__(self, "is_sorted", True)

        object.__setattr__(self, "items", tuple(unique_items))

    @classmethod
    def from_values(cls, values: Iterable[Any], *, is_sorted: bool = False) -> Self:
        """Create a CompletionItem instance from arbitrary objects.

        :param values: the raw objects (e.g. strs, ints, Paths) to be converted into CompletionItems.
        :param is_sorted: whether the values are already in the desired order.
        """
        items = [v if isinstance(v, CompletionItem) else CompletionItem(value=v) for v in values]
        return cls(items=items, is_sorted=is_sorted)

    def to_strings(self) -> tuple[str, ...]:
        """Return a tuple of the completion strings (the 'text' field of each item)."""
        return tuple(item.text for item in self.items)

    # --- Sequence Protocol Functions ---

    def __bool__(self) -> bool:
        """Return True if there are items, False otherwise."""
        return bool(self.items)

    def __len__(self) -> int:
        """Return the number of items."""
        return len(self.items)

    def __contains__(self, item: object) -> bool:
        """Return True if the item is present in the collection."""
        return item in self.items

    def __iter__(self) -> Iterator[CompletionItem]:
        """Allow the collection to be used in loops or comprehensions."""
        return iter(self.items)

    def __reversed__(self) -> Iterator[CompletionItem]:
        """Allow the collection to be iterated in reverse order using reversed()."""
        return reversed(self.items)

    @overload
    def __getitem__(self, index: int) -> CompletionItem: ...

    @overload
    def __getitem__(self, index: slice) -> tuple[CompletionItem, ...]: ...

    def __getitem__(self, index: int | slice) -> CompletionItem | tuple[CompletionItem, ...]:
        """Retrieve an item by its integer index or a range of items using a slice."""
        items_tuple = cast(tuple[CompletionItem, ...], self.items)
        return items_tuple[index]


@dataclass(frozen=True, slots=True, kw_only=True)
class Choices(CompletionResultsBase):
    """A collection of potential values available for completion, typically provided by a choice provider."""


@dataclass(frozen=True, slots=True, kw_only=True)
class Completions(CompletionResultsBase):
    """The results of a completion operation."""

    # Optional hint which prints above completion suggestions
    hint: str = ""

    # Optional message to display if an error occurs during completion
    error: str = ""

    # Optional Rich table which provides more context for the data being completed
    table: Table | None = None

    # If True, the completion engine is allowed to finalize a completion
    # when a single match is found by appending a trailing space and
    # closing any open quotation marks.
    #
    # Set this to False for intermediate or hierarchical matches (such as
    # directories) where the user needs to continue typing the next segment.
    # This flag is ignored if there are multiple matches.
    allow_finalization: bool = True

    #####################################################################
    # The following fields are used internally by cmd2 to handle
    # automatic quoting and are not intended for user modification.
    #####################################################################

    # Whether to add an opening quote to the matches.
    _add_opening_quote: bool = False

    # The starting index of the user-provided search text within a full match.
    # This accounts for leading shortcuts (e.g., in '?cmd', the offset is 1).
    # Used to ensure opening quotes are inserted after the shortcut rather than before it.
    _search_text_offset: int = 0

    # The quote character to use if adding an opening or closing quote to the matches.
    _quote_char: str = ""



"""Module adds capabilities to argparse by patching a few of its functions.

It also defines a parser class called Cmd2ArgumentParser which improves error
and help output over normal argparse. All cmd2 code uses this parser and it is
required that developers of cmd2-based apps either use it or write their own
parser that inherits from it. If you wish to override the parser used by cmd2's
built-in commands, see custom_parser.py example.


**Added capabilities**

Extends argparse nargs functionality by allowing tuples which specify a range
(min, max). To specify a max value with no upper bound, use a 1-item tuple
(min,)

Example::

    # -f argument expects at least 3 values
    parser.add_argument("-f", nargs=(3,))

    # -f argument expects 3 to 5 values
    parser.add_argument("-f", nargs=(3, 5))


**Completion**

cmd2 uses its ArgparseCompleter class to enable argparse-based completion
on all commands that use the @with_argparser decorator. Out of the box you get
completion of commands, subcommands, and flag names, as well as instructive
hints about the current argument that print when tab is pressed. In addition,
you can add completion for each argument's values using parameters passed
to add_argument().

Below are the 3 add_argument() parameters for enabling completion of an
argument's value. Only one can be used at a time.

``choices`` - pass a list of values to the choices parameter.

    Example::

        my_list = ["An Option", "SomeOtherOption"]
        parser.add_argument("-o", "--options", choices=my_list)

``choices_provider`` - pass a function that returns a Choices object. This is good in
cases where the choices are dynamically generated when the user hits tab.

    Example::

        def my_choices_provider(self) -> Choices:
            ...
            return my_choices


        parser.add_argument("arg", choices_provider=my_choices_provider)

``completer`` - pass a function that does custom completion and returns a Completions object.

cmd2 provides a few completer methods for convenience (e.g., path_complete,
delimiter_complete)

    Example::

        # This adds file-path completion to an argument
        parser.add_argument("-o", "--options", completer=cmd2.Cmd.path_complete)

    You can use functools.partial() to prepopulate values of the underlying
    choices and completer functions/methods.

    Example::

        # This says to call path_complete with a preset value for its path_filter argument
        dir_completer = functools.partial(path_complete, path_filter=lambda path: os.path.isdir(path))
        parser.add_argument("-o", "--options", completer=dir_completer)

For ``choices_provider`` and ``completer``, do not set them to a bound method. This
is because ArgparseCompleter passes the `self` argument explicitly to these
functions. When ArgparseCompleter calls one, it will detect whether it is bound
to a `Cmd` subclass or `CommandSet`. If bound to a `cmd2.Cmd subclass`, it will
pass the app instance as the `self` argument. If bound to a `cmd2.CommandSet`
subclass, it will pass the `CommandSet` instance as the `self` argument.
Therefore instead of passing something like `self.path_complete`, pass
`cmd2.Cmd.path_complete`.

``choices_provider`` and ``completer`` functions can also be implemented as
standalone functions (i.e. not a member of a class). In this case,
ArgparseCompleter will pass its ``cmd2.Cmd`` app instance as the first
positional argument.

Of the 3 completion parameters, ``choices`` is the only one where argparse
validates user input against items in the choices list. This is because the
other 2 parameters are meant to complete data sets that are viewed as
dynamic. Therefore it is up to the developer to validate if the user has typed
an acceptable value for these arguments.

There are times when what's being completed is determined by a previous
argument on the command line. In these cases, ArgparseCompleter can pass a
dictionary that maps the command line tokens up through the one being completed
to their argparse argument name. To receive this dictionary, your
choices/completer function should have an argument called arg_tokens.

    Example::

        def my_choices_provider(self, arg_tokens) -> Choices
        def my_completer(self, text, line, begidx, endidx, arg_tokens) -> Completions

All values of the arg_tokens dictionary are lists, even if a particular
argument expects only 1 token. Since ArgparseCompleter is for completion,
it does not convert the tokens to their actual argument types or validate their
values. All tokens are stored in the dictionary as the raw strings provided on
the command line. It is up to the developer to determine if the user entered
the correct argument type (e.g. int) and validate their values.

**CompletionItem Class**

This class represents a single completion result and what the ``Choices``
and ``Completion`` classes contain.

``CompletionItem`` provides the following optional metadata fields which enhance
completion results displayed to the screen.

1. display - string for displaying the completion differently in the completion menu
2. display_meta - meta information about completion which displays in the completion menu
3. table_data - supplemental data for completion tables

They can also be used as argparse choices. When a ``CompletionItem`` is created, it
stores the original value (e.g. ID number) and makes it accessible through a property
called ``value``. cmd2 has patched argparse so that when evaluating choices, input
is compared to ``CompletionItem.value`` instead of the ``CompletionItem`` instance.

**Completion Tables**

These were added to help in cases where uninformative data is being completed.
For instance, completing ID numbers isn't very helpful to a user without context.

Providing ``table_data`` in your ``CompletionItem`` signals ArgparseCompleter
to output the completion results in a table with supplemental data instead of just a table
of tokens::

    Instead of this:
        1     2     3

    The user sees this:
         ITEM_ID   Description
        ────────────────────────────
               1   My item
               2   Another item
               3   Yet another item


The left-most column is the actual value being completed and its header is
that value's name. Any additional column headers are defined using the
``table_columns`` parameter of add_argument(), which is a list of header
names. The supplemental column values come from the
``table_data`` argument to ``CompletionItem``. It's a ``Sequence`` with the
same number of items as ``table_columns``.

Example::

    Add an argument and define its table_columns.

        parser.add_argument(
            "item_id",
            type=int,
            choices_provider=get_choices,
            table_columns=["Item Name", "Checked Out", "Due Date"],
        )

    Implement the choices_provider to return Choices.

        def get_choices(self) -> Choices:
            \"\"\"choices_provider which returns CompletionItems\"\"\"

            # Populate CompletionItem's table_data argument.
            # Its item count should match that of table_columns.
            items = [
                CompletionItem(1, table_data=["My item", True, "02/02/2022"]),
                CompletionItem(2, table_data=["Another item", False, ""]),
                CompletionItem(3, table_data=["Yet another item", False, ""]),
            ]
            return Choices(items)

    This is what the user will see during completion.

        ITEM_ID   Item Name          Checked Out   Due Date
        ───────────────────────────────────────────────────────
              1   My item            True          02/02/2022
              2   Another item       False
              3   Yet another item   False

``table_columns`` can be strings or ``Rich.table.Columns`` for more
control over things like alignment.

- If a header is a string, it will render as a left-aligned column with its
overflow behavior set to "fold". This means a long string will wrap within its
cell, creating as many new lines as required to fit.

- If a header is a ``Column``, it defaults to "ellipsis" overflow behavior.
This means a long string which exceeds the width of its column will be
truncated with an ellipsis at the end. You can override this and other settings
when you create the ``Column``.

``table_data`` items can include Rich objects, including styled Text and Tables.

To avoid printing excessive information to the screen at once when a user
presses tab, there is a maximum threshold for the number of ``CompletionItems``
that will be shown. Its value is defined in ``cmd2.Cmd.max_completion_table_items``.
It defaults to 50, but can be changed. If the number of completion suggestions
exceeds this number, then a completion table won't be displayed.


**Custom Argument Parameters**

``argparse._ActionsContainer.add_argument`` has been patched to support several
custom parameters used for tab completion and nargs range parsing. These
parameters are registered using ``register_argparse_argument_parameter()``.
See ``_ActionsContainer_add_argument`` for more details on these parameters.

Registering a parameter whitelists it for use in ``add_argument()`` and
automatically adds getter and setter accessor methods to the ``argparse.Action``
class. For any registered parameter named ``<name>``, the following methods are
available on the resulting ``Action`` object to access its underlying attribute:

- ``action.get_<name>()``
- ``action.set_<name>(value)``
"""

import argparse
import contextlib
import re
import sys
import threading
from argparse import ArgumentError
from collections.abc import (
    Callable,
    Iterable,
    Iterator,
    Sequence,
)
from dataclasses import dataclass
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    NoReturn,
    TypeAlias,
    Union,
    cast,
)

from rich.table import Column

from . import constants
from .completion import CompletionItem
from .rich_utils import (
    Cmd2HelpFormatter,
    HelpContent,
)
from .styles import Cmd2Style
from .types import (
    CmdOrSetT,
    UnboundChoicesProvider,
    UnboundCompleter,
)

if TYPE_CHECKING:  # pragma: no cover
    from .argparse_completer import ArgparseCompleter

    # In Python 3.14+, move these definitions outside the TYPE_CHECKING
    # block as staticmethod/classmethod become subscriptable at runtime.
    _StaticParserFactory = staticmethod[[], "Cmd2ArgumentParser"]
    _ClassParserFactory = classmethod[CmdOrSetT, [], "Cmd2ArgumentParser"]
else:
    _StaticParserFactory = staticmethod
    _ClassParserFactory = classmethod

# Represents a parser factory with no arguments (including staticmethod)
NoParamParserFactory: TypeAlias = Callable[[], "Cmd2ArgumentParser"] | _StaticParserFactory

# Represents a parser factory with a class argument (including classmethod)
ClassParamParserFactory: TypeAlias = Union[
    Callable[[type[CmdOrSetT]], "Cmd2ArgumentParser"],
    "_ClassParserFactory[CmdOrSetT]",
]

# Represents the various types from which cmd2 can build a parser
ParserSource: TypeAlias = Union[
    "Cmd2ArgumentParser",
    NoParamParserFactory,
    ClassParamParserFactory[CmdOrSetT],
]


@dataclass(kw_only=True)
class ArgparseCommandSpec:
    """Metadata for an argparse-based command function.

    :param parser_source: an existing Cmd2ArgumentParser instance or a factory
                          (callable, staticmethod, or classmethod) that returns one.
    :param preserve_quotes: if True, then arguments passed to argparse maintain their quotes
    """

    parser_source: ParserSource[Any]
    preserve_quotes: bool = False


@dataclass(kw_only=True)
class _SubcommandBase:
    """Base metadata shared by all subcommand representations.

    :param name: the name of the subcommand
    :param command: the full parent command path (e.g., 'foo bar')
    :param help: optional help message for this subcommand
    :param aliases: optional alternative names for this subcommand
    :param deprecated: whether this subcommand is deprecated (requires Python 3.13+).
    """

    name: str
    command: str
    help: str | None = None
    aliases: tuple[str, ...] = ()
    deprecated: bool = False


@dataclass(kw_only=True)
class SubcommandSpec(_SubcommandBase):
    """Metadata used to build and register a subcommand.

    :param parser_source: an existing Cmd2ArgumentParser instance or a factory
                          (callable, staticmethod, or classmethod) that returns one.
    """

    parser_source: ParserSource[Any]


@dataclass(kw_only=True)
class SubcommandRecord(_SubcommandBase):
    """A record of a subcommand's configuration and parser.

    Used primarily for attaching and detaching subcommands.

    :param parser: the built Cmd2ArgumentParser instance for this subcommand
    """

    parser: "Cmd2ArgumentParser"


def build_range_error(range_min: int, range_max: float) -> str:
    """Build an error message when the number of arguments provided is not within the expected range."""
    err_msg = "expected "

    if range_max == constants.INFINITY:
        plural = "" if range_min == 1 else "s"
        err_msg += f"at least {range_min}"
    else:
        plural = "" if range_max == 1 else "s"
        if range_min == range_max:
            err_msg += f"{range_min}"
        else:
            err_msg += f"{range_min} to {range_max}"

    err_msg += f" argument{plural}"

    return err_msg


############################################################################################################
# Allow developers to add custom action attributes
############################################################################################################

# This set should only be edited by calling register_argparse_argument_parameter().
# Do not manually add or remove items.
_CUSTOM_ACTION_ATTRIBS: set[str] = set()


def register_argparse_argument_parameter(
    param_name: str,
    *,
    validator: Callable[[argparse.Action, Any], Any] | None = None,
) -> None:
    """Register a custom parameter for argparse.Action and add accessors to the Action class.

    :param param_name: Name of the parameter. This must be a valid Python identifier.
    :param validator: Optional function to validate and/or transform the parameter value.
                      It accepts the Action instance and the value as arguments.
    :raises ValueError: if the parameter name is invalid
    :raises KeyError: if the new parameter collides with any existing attributes
    """
    if not param_name.isidentifier():
        raise ValueError(f"Invalid parameter name '{param_name}': must be a valid Python identifier")

    if param_name in _CUSTOM_ACTION_ATTRIBS:
        raise KeyError(f"Custom parameter '{param_name}' is already registered")

    # Ensure we don't hijack standard argparse.Action attributes or existing methods
    if hasattr(argparse.Action, param_name):
        raise KeyError(f"'{param_name}' conflicts with an existing attribute on argparse.Action")

    # Check if accessors already exist (e.g., from manual patching or previous registration)
    getter_name = f"get_{param_name}"
    setter_name = f"set_{param_name}"
    if hasattr(argparse.Action, getter_name) or hasattr(argparse.Action, setter_name):
        raise KeyError(f"Accessor methods for '{param_name}' already exist on argparse.Action")

    # Check for the prefixed internal attribute name collision (e.g., _cmd2_<param_name>)
    attr_name = constants.cmd2_private_attr_name(param_name)
    if hasattr(argparse.Action, attr_name):
        raise KeyError(f"The internal attribute '{attr_name}' already exists on argparse.Action")

    def _action_get_custom_parameter(self: argparse.Action) -> Any:
        """Get the custom attribute of an argparse Action."""
        return getattr(self, attr_name, None)

    setattr(argparse.Action, getter_name, _action_get_custom_parameter)

    def _action_set_custom_parameter(self: argparse.Action, value: Any) -> None:
        """Set the custom attribute of an argparse Action."""
        if validator is not None:
            value = validator(self, value)

        setattr(self, attr_name, value)

    setattr(argparse.Action, setter_name, _action_set_custom_parameter)

    _CUSTOM_ACTION_ATTRIBS.add(param_name)


def _validate_completion_callable(self: argparse.Action, value: Any) -> Any:
    """Validate choices_provider and completer values for potential conflicts."""
    if value is None:
        return None

    if self.choices is not None:
        err_msg = "None of the following parameters can be used alongside a choices parameter:\nchoices_provider, completer"
        raise ValueError(err_msg)
    if self.nargs == 0:
        err_msg = (
            "None of the following parameters can be used on an action that takes no arguments:\nchoices_provider, completer"
        )
        raise ValueError(err_msg)
    return value


# Add new attributes to argparse.Action.
# See _ActionsContainer_add_argument() for details on these attributes.
register_argparse_argument_parameter("choices_provider", validator=_validate_completion_callable)
register_argparse_argument_parameter("completer", validator=_validate_completion_callable)
register_argparse_argument_parameter("table_columns")
register_argparse_argument_parameter("nargs_range")
register_argparse_argument_parameter("suppress_tab_hint")


############################################################################################################
# Patch _ActionsContainer.add_argument to support more arguments
############################################################################################################

# Save original _ActionsContainer.add_argument so we can call it in our patch
orig_actions_container_add_argument = argparse._ActionsContainer.add_argument


def _ActionsContainer_add_argument(  # noqa: N802
    self: argparse._ActionsContainer,
    *args: Any,
    nargs: int | str | tuple[int] | tuple[int, int] | tuple[int, float] | None = None,
    choices_provider: UnboundChoicesProvider[CmdOrSetT] | None = None,
    completer: UnboundCompleter[CmdOrSetT] | None = None,
    suppress_tab_hint: bool = False,
    table_columns: Sequence[str | Column] | None = None,
    **kwargs: Any,
) -> argparse.Action:
    """Patch _ActionsContainer.add_argument() to support cmd2-specific settings.

    # Args from original function
    :param self: instance of the _ActionsContainer being added to
    :param args: arguments expected by argparse._ActionsContainer.add_argument

    # Customized arguments from original function
    :param nargs: extends argparse nargs functionality by allowing tuples which specify a range (min, max)
                  to specify a max value with no upper bound, use a 1-item tuple (min,)

    # Added args used by ArgparseCompleter
    :param choices_provider: function that provides choices for this argument
    :param completer: completion function that provides choices for this argument
    :param suppress_tab_hint: when ArgparseCompleter has no results to show during completion, it displays the
                              current argument's help text as a hint. Set this to True to suppress the hint. If this
                              argument's help text is set to argparse.SUPPRESS, then tab hints will not display
                              regardless of the value passed for suppress_tab_hint. Defaults to False.
    :param table_columns: optional headers for when displaying a completion table. Defaults to None.

    # Args from original function
    :param kwargs: keyword-arguments recognized by argparse._ActionsContainer.add_argument

    Note: You can only use 1 of the following in your argument:
          choices, choices_provider, completer

          See the header of this file for more information

    :return: the created argument action
    :raises ValueError: on incorrect parameter usage
    """
    # Verify consistent use of arguments
    if choices_provider is not None and completer is not None:
        raise ValueError("Only one of the following parameters may be used at a time:\nchoices_provider, completer")

    # Pre-process special ranged nargs
    nargs_range = None

    if nargs is not None:
        nargs_adjusted: int | str | tuple[int] | tuple[int, int] | tuple[int, float] | None
        # Check if nargs was given as a range
        if isinstance(nargs, tuple):
            # Handle 1-item tuple by setting max to INFINITY
            if len(nargs) == 1:
                nargs = (nargs[0], constants.INFINITY)

            # Validate nargs tuple
            if (
                len(nargs) != 2
                or not isinstance(nargs[0], int)
                or not (isinstance(nargs[1], int) or nargs[1] == constants.INFINITY)
            ):
                raise ValueError("Ranged values for nargs must be a tuple of 1 or 2 integers")
            if nargs[0] >= nargs[1]:
                raise ValueError("Invalid nargs range. The first value must be less than the second")
            if nargs[0] < 0:
                raise ValueError("Negative numbers are invalid for nargs range")

            # Save the nargs tuple as our range setting
            nargs_range = nargs
            range_min = nargs_range[0]
            range_max = nargs_range[1]

            # Convert nargs into a format argparse recognizes
            if range_min == 0:
                if range_max == 1:
                    nargs_adjusted = argparse.OPTIONAL

                    # No range needed since (0, 1) is just argparse.OPTIONAL
                    nargs_range = None
                else:
                    nargs_adjusted = argparse.ZERO_OR_MORE
                    if range_max == constants.INFINITY:
                        # No range needed since (0, INFINITY) is just argparse.ZERO_OR_MORE
                        nargs_range = None
            elif range_min == 1 and range_max == constants.INFINITY:
                nargs_adjusted = argparse.ONE_OR_MORE

                # No range needed since (1, INFINITY) is just argparse.ONE_OR_MORE
                nargs_range = None
            else:
                nargs_adjusted = argparse.ONE_OR_MORE
        else:
            nargs_adjusted = nargs

        # Add the argparse-recognized version of nargs to kwargs
        kwargs["nargs"] = nargs_adjusted

    # Extract registered custom keyword arguments
    custom_attribs = {keyword: value for keyword, value in kwargs.items() if keyword in _CUSTOM_ACTION_ATTRIBS}
    for keyword in custom_attribs:
        del kwargs[keyword]

    # Create the argument using the original add_argument function
    new_arg = orig_actions_container_add_argument(self, *args, **kwargs)

    # Set the cmd2-specific attributes
    new_arg.set_nargs_range(nargs_range)  # type: ignore[attr-defined]
    new_arg.set_choices_provider(choices_provider)  # type: ignore[attr-defined]
    new_arg.set_completer(completer)  # type: ignore[attr-defined]
    new_arg.set_suppress_tab_hint(suppress_tab_hint)  # type: ignore[attr-defined]
    new_arg.set_table_columns(table_columns)  # type: ignore[attr-defined]

    # Set other registered custom attributes
    for keyword, value in custom_attribs.items():
        attr_setter = getattr(new_arg, f"set_{keyword}", None)
        if attr_setter is not None:
            attr_setter(value)

    return new_arg


# Overwrite _ActionsContainer.add_argument with our patch
argparse._ActionsContainer.add_argument = _ActionsContainer_add_argument  # type: ignore[method-assign]

############################################################################################################
# Patch argparse._SubParsersAction by adding remove_parser() function
############################################################################################################


def _SubParsersAction_remove_parser(  # noqa: N802
    self: "argparse._SubParsersAction[Cmd2ArgumentParser]",
    name: str,
) -> SubcommandRecord:
    """Remove a subparser from a subparsers group.

    This function is added by cmd2 as a method called ``remove_parser()``
    to ``argparse._SubParsersAction`` class.

    To call: ``action.remove_parser(name)``

    :param self: instance of the _SubParsersAction being edited
    :param name: name of the subcommand for the subparser to remove
    :return: a SubcommandRecord object describing the removed parser
    :raises ValueError: if the subcommand doesn't exist
    """
    if name not in self._name_parser_map:
        raise ValueError(f"Subcommand '{name}' does not exist")

    subparser = self._name_parser_map[name]

    # Find all names (primary and aliases) that map to this subparser
    all_names = [cur_name for cur_name, cur_parser in self._name_parser_map.items() if cur_parser is subparser]

    # argparse inserts the primary name before the aliases in _name_parser_map
    primary_name = all_names[0]
    aliases = tuple(all_names[1:])

    # Handle Python 3.13+ deprecation
    deprecated: bool = False
    deprecated_attr = getattr(self, "_deprecated", None)
    if isinstance(deprecated_attr, set):
        if primary_name in deprecated_attr:
            deprecated = True
            deprecated_attr.discard(primary_name)
        for alias in aliases:
            deprecated_attr.discard(alias)

    # Remove the help entry for this subparser.
    help_text = None
    for choice_action in self._choices_actions:
        if choice_action.dest == primary_name:
            help_text = choice_action.help
            self._choices_actions.remove(choice_action)
            break

    # Remove all references to this subparser, including aliases.
    for cur_name in all_names:
        del self._name_parser_map[cur_name]

    return SubcommandRecord(
        name=primary_name,
        command="",  # To be populated by the caller
        help=help_text,
        aliases=aliases,
        deprecated=deprecated,
        parser=subparser,
    )


def _SubParsersAction_remove_all_parsers(  # noqa: N802
    self: "argparse._SubParsersAction[Cmd2ArgumentParser]",
) -> list[SubcommandRecord]:
    """Remove all subparsers from a subparsers group.

    This function is added by cmd2 as a method called ``remove_all_parsers()``
    to ``argparse._SubParsersAction`` class.

    To call: ``action.remove_all_parsers()``

    :param self: instance of the _SubParsersAction being edited
    :return: a list of SubcommandRecord objects for the removed subparsers
    """
    records: list[SubcommandRecord] = []

    while self._name_parser_map:
        # Get the next subcommand name. remove_parser() will remove
        # it and any associated aliases from _name_parser_map.
        name = next(iter(self._name_parser_map))
        record = self.remove_parser(name)  # type: ignore[attr-defined]
        records.append(record)

    return records


argparse._SubParsersAction.remove_parser = _SubParsersAction_remove_parser  # type: ignore[attr-defined]
argparse._SubParsersAction.remove_all_parsers = _SubParsersAction_remove_all_parsers  # type: ignore[attr-defined]


@dataclass
class _ParserThreadLocals(threading.local):
    """Thread-local storage used by Cmd2ArgumentParser to manage execution context."""

    # The active output stream for help, usage, and errors. Since argparse does not
    # pass the destination stream to the formatter factory, this transient value
    # provides the context needed to synchronize Rich's rendering with the specific
    # capabilities of the destination file descriptor. It is managed via the
    # output_to() context manager.
    current_output_file: IO[str] | None = None


class Cmd2ArgumentParser(argparse.ArgumentParser):
    """Custom ArgumentParser class that improves error and help output."""

    # Thread-local storage shared by all parser instances (including subparsers)
    _thread_locals: ClassVar[_ParserThreadLocals] = _ParserThreadLocals()

    @contextlib.contextmanager
    def output_to(self, file: IO[str] | None) -> Iterator[None]:
        """Context manager to temporarily set the output stream during argparse operations.

        This is helpful for directing output for functions like `parse_args()`, which
        default to `sys.stdout` and lack a `file` argument.

        :param file: the file stream to use for output
        """
        previous = self._thread_locals.current_output_file
        self._thread_locals.current_output_file = file
        try:
            yield
        finally:
            self._thread_locals.current_output_file = previous

    def __init__(
        self,
        prog: str | None = None,
        usage: str | None = None,
        description: HelpContent | None = None,
        epilog: HelpContent | None = None,
        parents: Sequence[argparse.ArgumentParser] = (),
        formatter_class: type[Cmd2HelpFormatter] = Cmd2HelpFormatter,
        prefix_chars: str = "-",
        fromfile_prefix_chars: str | None = None,
        argument_default: str | None = None,
        conflict_handler: str = "error",
        add_help: bool = True,
        allow_abbrev: bool = True,
        exit_on_error: bool = True,
        suggest_on_error: bool = False,
        color: bool = False,
        *,
        completer_class: type["ArgparseCompleter"] | None = None,
    ) -> None:
        """Initialize the Cmd2ArgumentParser instance.

        :param completer_class: optional parameter which specifies a subclass of ArgparseCompleter
                                for custom completion behavior on this parser. If this is None, then
                                it will be set to argparse_completer.DEFAULT_ARGPARSE_COMPLETER.
        """
        kwargs: dict[str, bool] = {}
        if sys.version_info >= (3, 14):
            # Python >= 3.14 so pass new arguments to parent argparse.ArgumentParser class
            kwargs = {
                "suggest_on_error": suggest_on_error,
                "color": color,
            }

        super().__init__(
            prog=prog,
            usage=usage,
            description=description,  # type: ignore[arg-type, ty:invalid-argument-type]
            epilog=epilog,  # type: ignore[arg-type, ty:invalid-argument-type]
            parents=parents,
            formatter_class=formatter_class,
            prefix_chars=prefix_chars,
            fromfile_prefix_chars=fromfile_prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
            allow_abbrev=allow_abbrev,
            exit_on_error=exit_on_error,
            **kwargs,
        )

        if completer_class is None:
            from . import argparse_completer

            completer_class = argparse_completer.DEFAULT_ARGPARSE_COMPLETER
        self.completer_class = completer_class

        # To assist type checkers, recast these to reflect our usage of rich-argparse.
        self.formatter_class: type[Cmd2HelpFormatter]
        self.description: HelpContent | None  # type: ignore[assignment]
        self.epilog: HelpContent | None  # type: ignore[assignment]

    def print_usage(self, file: IO[str] | None = None) -> None:  # type: ignore[override, ty:invalid-method-override]
        """Override to ensure the formatter is aware of the target file."""
        if file is None:
            file = self._thread_locals.current_output_file

        with self.output_to(file):
            super().print_usage(file)

    def print_help(self, file: IO[str] | None = None) -> None:  # type: ignore[override, ty:invalid-method-override]
        """Override to ensure the formatter is aware of the target file."""
        if file is None:
            file = self._thread_locals.current_output_file

        with self.output_to(file):
            super().print_help(file)

    def get_subparsers_action(self) -> "argparse._SubParsersAction[Cmd2ArgumentParser]":
        """Get the _SubParsersAction for this parser if it exists.

        :return: the _SubParsersAction for this parser
        :raises ValueError: if this parser does not support subcommands
        """
        if self._subparsers is not None:
            for action in self._subparsers._group_actions:
                if isinstance(action, argparse._SubParsersAction):
                    return action
        raise ValueError(f"Command '{self.prog}' does not support subcommands")

    def _build_subparsers_prog_prefix(self, positionals: list[argparse.Action]) -> str:
        """Build the 'prog' prefix for a subparsers action.

        This prefix is stored in the _SubParsersAction's '_prog_prefix' attribute and
        is used to construct the 'prog' attribute for its child parsers. It
        typically consists of the current parser's 'prog' name followed by any
        positional arguments that appear before the _SubParsersAction.

        This method uses a temporary Cmd2ArgumentParser to leverage argparse's
        functionality for generating these strings. Subclasses can override this if
        they need to change how subcommand 'prog' values are constructed (e.g., if
        add_subparsers() was overridden with custom naming logic or if a different
        formatting style is desired).

        Note: This method explicitly instantiates Cmd2ArgumentParser rather than
        type(self) to avoid potential side effects or mandatory constructor
        arguments in user-defined subclasses.

        :param positionals: positional arguments which appear before the _SubParsersAction
        :return: the built 'prog' prefix
        """
        # 1. usage=None: In Python < 3.14, this prevents the default usage
        #    string from affecting subparser prog strings. This was fixed in 3.14:
        #    https://github.com/python/cpython/commit/0cb4d6c6549d2299f7518f083bbe7d10314ecd66
        #
        # 2. add_help=False: No need for a help action since we already know which
        #    actions are needed to build the prefix and have passed them in
        #    via the 'positionals' argument.
        temp_parser = Cmd2ArgumentParser(
            prog=self.prog,
            usage=None,
            formatter_class=cast(type[Cmd2HelpFormatter], self.formatter_class),
            add_help=False,
        )

        # Inject the current positional state so add_subparsers() has the right context
        temp_parser._actions = positionals
        temp_parser._mutually_exclusive_groups = self._mutually_exclusive_groups

        # Call add_subparsers() to build _prog_prefix
        return temp_parser.add_subparsers()._prog_prefix

    def update_prog(self, prog: str) -> None:
        """Recursively update the prog attribute of this parser and all of its subparsers.

        :param prog: new value for this parser's prog attribute
        """
        # Set the prog value for this parser
        self.prog = prog

        try:
            subparsers_action = self.get_subparsers_action()
        except ValueError:
            # This parser has no subcommands
            return

        # Get all positional arguments which appear before the subcommand.
        positionals: list[argparse.Action] = []
        for action in self._actions:
            if action is subparsers_action:
                break

            # Save positional argument
            if not action.option_strings:
                positionals.append(action)

        # Update _prog_prefix. This ensures that any subcommands added later via
        # add_parser() will have the correct prog value.
        subparsers_action._prog_prefix = self._build_subparsers_prog_prefix(positionals)

        # subparsers_action._name_parser_map includes aliases. Since primary names are inserted
        # first, we skip already updated parsers to ensure primary names are used in 'prog'.
        # We can't rely on subparsers_action._choices_actions to filter out aliases because while
        # it contains only primary names, it omits any subcommands that lack help text.
        updated_parsers: set[Cmd2ArgumentParser] = set()

        # Set the prog value for each subcommand's parser
        for subcmd_name, subcmd_parser in subparsers_action._name_parser_map.items():
            if subcmd_parser in updated_parsers:
                continue

            subcmd_prog = f"{subparsers_action._prog_prefix} {subcmd_name}"
            subcmd_parser.update_prog(subcmd_prog)
            updated_parsers.add(subcmd_parser)

    def find_parser(self, subcommand_path: Iterable[str]) -> "Cmd2ArgumentParser":
        """Find a parser in the hierarchy based on a sequence of subcommand names.

        :param subcommand_path: sequence of subcommand names leading to the target parser
        :return: the discovered parser
        :raises ValueError: if any subcommand in the path is not found or a level doesn't support subcommands
        """
        parser = self
        for name in subcommand_path:
            subparsers_action = parser.get_subparsers_action()
            if name not in subparsers_action._name_parser_map:
                raise ValueError(f"Subcommand '{name}' does not exist for '{parser.prog}'")
            parser = subparsers_action._name_parser_map[name]
        return parser

    def attach_subcommand(
        self,
        record: SubcommandRecord,
        subcommand_path: Iterable[str] = (),
    ) -> None:
        """Attach a parser as a subcommand to a command at the specified path.

        Note: `record.command` is not used for navigation here. It is assumed you
        are attaching relative to `self` using `subcommand_path`. However,
        `record.command` will be updated to reflect the final, absolute path
        of the parent parser this subcommand is attached to.

        :param record: SubcommandRecord object describing the subcommand
        :param subcommand_path: sequence of subcommand names leading to the parser that will
                                host the new subcommand. An empty sequence indicates this parser.
        :raises TypeError: if record.parser is not an instance of Cmd2ArgumentParser (or subclass)
        :raises ValueError: if the command path is invalid, doesn't support subcommands, or the
                            subcommand already exists
        """
        if not isinstance(record.parser, Cmd2ArgumentParser):
            raise TypeError(
                f"The attached parser must be an instance of 'Cmd2ArgumentParser' (or subclass). "
                f"Received: '{type(record.parser).__name__}'."
            )

        target_parser = self.find_parser(subcommand_path)
        subparsers_action = target_parser.get_subparsers_action()

        # Verify the parser is compatible with the 'parser_class' configured for this
        # subcommand group. We use isinstance() here to allow for subclasses, providing
        # more flexibility than the standard add_parser() factory approach which enforces
        # a specific class.
        if not isinstance(record.parser, subparsers_action._parser_class):
            raise TypeError(
                f"The attached parser must be an instance of '{subparsers_action._parser_class.__name__}' "
                f"(or subclass) to match the 'parser_class' configured for this subcommand group. "
                f"Received: '{type(record.parser).__name__}'."
            )

        # Do not overwrite existing subcommands or aliases
        all_names = (record.name, *record.aliases)
        for name in all_names:
            if name in subparsers_action._name_parser_map:
                raise ValueError(f"Subcommand '{name}' already exists for '{target_parser.prog}'")

        # Registration kwargs
        kwargs: dict[str, Any] = {"aliases": record.aliases}
        if record.help is not None:
            kwargs["help"] = record.help
        if record.deprecated:
            kwargs["deprecated"] = record.deprecated

        # Use add_parser to register the subcommand name and any aliases
        placeholder_parser = subparsers_action.add_parser(record.name, **kwargs)

        # To ensure accurate usage strings, recursively update 'prog' values
        # within the injected parser to match its new location in the command hierarchy.
        record.parser.update_prog(placeholder_parser.prog)

        # Replace the parser created by add_parser() with our pre-configured one
        subparsers_action._name_parser_map[record.name] = record.parser

        # Remap any aliases to our pre-configured parser
        for alias in record.aliases:
            subparsers_action._name_parser_map[alias] = record.parser

        # Update command to reflect the parent parser's absolute path
        record.command = target_parser.prog

    def detach_subcommand(self, subcommand_path: Iterable[str], subcommand: str) -> SubcommandRecord:
        """Detach a subcommand from a command at the specified path.

        :param subcommand_path: sequence of subcommand names leading to the parser hosting the
                                subcommand to be detached. An empty sequence indicates this parser.
        :param subcommand: name of the subcommand to detach
        :return: a SubcommandRecord object describing the detached subcommand
        :raises ValueError: if the command path is invalid or the subcommand doesn't exist
        """
        target_parser = self.find_parser(subcommand_path)
        subparsers_action = target_parser.get_subparsers_action()

        try:
            record = cast(
                SubcommandRecord,
                subparsers_action.remove_parser(subcommand),  # type: ignore[attr-defined]
            )
        except ValueError:
            raise ValueError(f"Subcommand '{subcommand}' does not exist for '{target_parser.prog}'") from None

        # Update command to reflect the parent parser's absolute path
        record.command = target_parser.prog
        return record

    def detach_all_subcommands(self, subcommand_path: Iterable[str]) -> list[SubcommandRecord]:
        """Detach all subcommands from a command at the specified path.

        :param subcommand_path: sequence of subcommand names leading to the parser hosting the
                                subcommands to be detached. An empty sequence indicates this parser.
        :return: a list of SubcommandRecord objects describing the detached subcommands
        :raises ValueError: if the command path is invalid or the command doesn't support subcommands
        """
        target_parser = self.find_parser(subcommand_path)
        subparsers_action = target_parser.get_subparsers_action()

        records = cast(
            list[SubcommandRecord],
            subparsers_action.remove_all_parsers(),  # type: ignore[attr-defined]
        )
        # Update command for each detached subcommand
        for record in records:
            record.command = target_parser.prog
        return records

    def error(self, message: str) -> NoReturn:
        """Override that applies custom formatting to the error message."""
        lines = message.split("\n")
        formatted_message = ""
        for linum, line in enumerate(lines):
            if linum == 0:
                formatted_message = "Error: " + line
            else:
                formatted_message += "\n       " + line

        with self.output_to(sys.stderr):
            self.print_usage(sys.stderr)

            # Use console to add style since it will respect ALLOW_STYLE's value.
            # Now _get_formatter() will return a formatter bound to stderr.
            console = self._get_formatter().console
            with console.capture() as capture:
                console.print(formatted_message, style=Cmd2Style.ERROR)
            formatted_message = f"{capture.get()}"

        self.exit(2, f"{formatted_message}\n")

    def _get_formatter(self, *_args: Any, **_kwargs: Any) -> Cmd2HelpFormatter:
        """Override with customizations for Cmd2HelpFormatter."""
        formatter_class = cast(type[Cmd2HelpFormatter], self.formatter_class)
        return formatter_class(prog=self.prog, file=self._thread_locals.current_output_file)

    def format_help(self, *args: Any, **kwargs: Any) -> str:
        """Override to add a newline."""
        return super().format_help(*args, **kwargs) + "\n"

    def _get_nargs_pattern(self, action: argparse.Action) -> str:
        """Override to support nargs ranges."""
        nargs_range = action.get_nargs_range()  # type: ignore[attr-defined]
        if nargs_range:
            range_max = "" if nargs_range[1] == constants.INFINITY else nargs_range[1]
            nargs_pattern = f"(-*A{{{nargs_range[0]},{range_max}}}-*)"

            # if this is an optional action, -- is not allowed
            if action.option_strings:
                nargs_pattern = nargs_pattern.replace("-*", "")
                nargs_pattern = nargs_pattern.replace("-", "")
            return nargs_pattern

        return super()._get_nargs_pattern(action)

    def _match_argument(self, action: argparse.Action, arg_strings_pattern: str) -> int:
        """Override to support nargs ranges."""
        nargs_pattern = self._get_nargs_pattern(action)
        match = re.match(nargs_pattern, arg_strings_pattern)

        # raise an exception if we weren't able to find a match
        if match is None:
            nargs_range = action.get_nargs_range()  # type: ignore[attr-defined]
            if nargs_range is not None:
                raise ArgumentError(action, build_range_error(nargs_range[0], nargs_range[1]))

        return super()._match_argument(action, arg_strings_pattern)

    def _check_value(self, action: argparse.Action, value: Any) -> None:
        """Override that supports CompletionItems as choices.

        When displaying choices, use CompletionItem.value instead of the CompletionItem instance.

        :param action: the action being populated
        :param value: value from command line already run through conversion function by argparse
        """
        # Import gettext like argparse does
        from gettext import gettext as _

        if action.choices is not None and value not in action.choices:
            # If any choice is a CompletionItem, then display its value property.
            choices = [c.value if isinstance(c, CompletionItem) else c for c in action.choices]
            args = {"value": value, "choices": ", ".join(map(repr, choices))}
            msg = _("invalid choice: %(value)r (choose from %(choices)s)")
            raise ArgumentError(action, msg % args)


# Parser type used by cmd2's built-in commands.
# Set it using cmd2.set_default_argument_parser().
DEFAULT_ARGUMENT_PARSER: type[Cmd2ArgumentParser] = Cmd2ArgumentParser


def set_default_argument_parser(parser_class: type[Cmd2ArgumentParser]) -> None:
    """Set the default Cmd2ArgumentParser class for cmd2's built-in commands.

    Since built-in commands rely on customizations made in Cmd2ArgumentParser,
    your custom parser class should inherit from Cmd2ArgumentParser.

    This should be called prior to instantiating your CLI object.

    See examples/custom_parser.py.
    """
    global DEFAULT_ARGUMENT_PARSER  # noqa: PLW0603
    DEFAULT_ARGUMENT_PARSER = parser_class



"""Custom exceptions for cmd2."""

from typing import Any

############################################################################################################
# The following exceptions are part of the public API
############################################################################################################


class SkipPostcommandHooks(Exception):  # noqa: N818
    """For when a command has a failed bad enough to skip post command hooks, but not bad enough to print to the user."""


class Cmd2ArgparseError(SkipPostcommandHooks):
    """A ``SkipPostcommandHooks`` exception for when a command fails to parse its arguments.

    Normally argparse raises a SystemExit exception in these cases. To avoid stopping the command
    loop, catch the SystemExit and raise this instead. If you still need to run post command hooks
    after parsing fails, just return instead of raising an exception.
    """


class CommandSetRegistrationError(Exception):
    """For when an error occurs while a CommandSet is being added or removed from a cmd2 application."""


class CompletionError(Exception):
    """Raised during completion operations to report any sort of error you want printed.

    Example use cases:

    - Reading a database to retrieve a completion data set failed
    - A previous command line argument that determines the data set being completed is invalid
    """

    def __init__(self, *args: Any, apply_style: bool = True) -> None:
        """Initialize CompletionError instance.

        :param apply_style: If True, then styles.ERROR will be applied to the message text when printed.
                            Set to False in cases where the message text already has the desired style.
                            Defaults to True.
        """
        self.apply_style = apply_style

        super().__init__(*args)


class PassThroughException(Exception):  # noqa: N818
    """Normally all unhandled exceptions raised during commands get printed to the user.

    This class is used to wrap an exception that should be raised instead of printed.
    """

    def __init__(self, *args: Any, wrapped_ex: BaseException) -> None:
        """Initialize PassThroughException instance.

        :param wrapped_ex: the exception that will be raised.
        """
        self.wrapped_ex = wrapped_ex
        super().__init__(*args)


############################################################################################################
# The following exceptions are NOT part of the public API and are intended for internal use only.
############################################################################################################


class Cmd2ShlexError(Exception):
    """Raised when shlex fails to parse a command line string in StatementParser."""


class EmbeddedConsoleExit(SystemExit):
    """Custom exception class for use with the py command."""


class EmptyStatement(Exception):  # noqa: N818
    """Custom exception class for handling behavior when the user just presses <Enter>."""


class IncompleteStatement(Exception):  # noqa: N818
    """Raised when more input is required to complete a multiline statement."""


class MacroError(Exception):
    """Raised when a macro fails to resolve (e.g., insufficient arguments)."""


class RedirectionError(Exception):
    """Custom exception class for when redirecting or piping output fails."""



"""Module defines the ArgparseCompleter class which provides argparse-based completion to cmd2 apps.

See the header of argparse_utils.py for instructions on how to use these features.
"""

import argparse
import dataclasses
import inspect
from collections import deque
from collections.abc import (
    Mapping,
    MutableSequence,
    Sequence,
)
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    cast,
)

from rich.table import Column
from rich.text import Text

from .argparse_utils import (
    Cmd2ArgumentParser,
    build_range_error,
)
from .command_set import CommandSet
from .completion import (
    CompletionItem,
    Completions,
)
from .constants import INFINITY
from .exceptions import CompletionError
from .rich_utils import Cmd2SimpleTable
from .types import (
    CmdOrSetT,
    UnboundChoicesProvider,
    UnboundCompleter,
)

if TYPE_CHECKING:  # pragma: no cover
    from .cmd2 import Cmd


# Name of the choice/completer function argument that, if present, will be passed a dictionary of
# command line tokens up through the token being completed mapped to their argparse destination name.
ARG_TOKENS = "arg_tokens"


def _build_hint(parser: Cmd2ArgumentParser, arg_action: argparse.Action) -> str:
    """Build completion hint for a given argument."""
    # Check if hinting is disabled for this argument
    suppress_hint = arg_action.get_suppress_tab_hint()  # type: ignore[attr-defined]
    if suppress_hint or arg_action.help == argparse.SUPPRESS:
        return ""

    # Use the parser's help formatter to display just this action's help text
    formatter = parser._get_formatter()
    formatter.start_section("Hint")
    formatter.add_argument(arg_action)
    formatter.end_section()
    return formatter.format_help()


def _single_prefix_char(token: str, parser: Cmd2ArgumentParser) -> bool:
    """Is a token just a single flag prefix character."""
    return len(token) == 1 and token[0] in parser.prefix_chars


def _looks_like_flag(token: str, parser: Cmd2ArgumentParser) -> bool:
    """Determine if a token looks like a flag.

    Unless an argument has nargs set to argparse.REMAINDER, then anything that looks like a flag
    can't be consumed as a value for it.

    Based on argparse._parse_optional().
    """
    # Flags have to be at least characters
    if len(token) < 2:
        return False

    # Flags have to start with a prefix character
    if token[0] not in parser.prefix_chars:
        return False

    # If it looks like a negative number, it is not a flag unless there are negative-number-like flags
    if parser._negative_number_matcher.match(token) and not parser._has_negative_number_optionals:
        return False

    # Flags can't have a space
    return " " not in token


class _ArgumentState:
    """Keeps state of an argument being parsed."""

    def __init__(self, arg_action: argparse.Action) -> None:
        self.action = arg_action
        self.min: int
        self.max: float | int
        self.count = 0
        self.is_remainder = self.action.nargs == argparse.REMAINDER

        # Check if nargs is a range
        nargs_range: tuple[int, int | float] | None = self.action.get_nargs_range()  # type: ignore[attr-defined]
        if nargs_range is not None:
            self.min = nargs_range[0]
            self.max = nargs_range[1]

        # Otherwise check against argparse types
        elif self.action.nargs is None:
            self.min = 1
            self.max = 1
        elif self.action.nargs == argparse.OPTIONAL:
            self.min = 0
            self.max = 1
        elif self.action.nargs in (argparse.ZERO_OR_MORE, argparse.REMAINDER):
            self.min = 0
            self.max = INFINITY
        elif self.action.nargs == argparse.ONE_OR_MORE:
            self.min = 1
            self.max = INFINITY
        else:
            self.min = cast(int, self.action.nargs)
            self.max = cast(int, self.action.nargs)


class _UnfinishedFlagError(CompletionError):
    def __init__(self, flag_arg_state: _ArgumentState) -> None:
        """CompletionError which occurs when the user has not finished the current flag.

        :param flag_arg_state: information about the unfinished flag action.
        """
        arg = f"{argparse._get_action_name(flag_arg_state.action)}"
        err = f"{build_range_error(flag_arg_state.min, flag_arg_state.max)}"
        error = f"Error: argument {arg}: {err} ({flag_arg_state.count} entered)"
        super().__init__(error)


class _NoResultsError(CompletionError):
    def __init__(self, parser: Cmd2ArgumentParser, arg_action: argparse.Action) -> None:
        """CompletionError which occurs when there are no results.

        If hinting is allowed on this argument, then its hint text will display.

        :param parser: Cmd2ArgumentParser instance which owns the action being completed
        :param arg_action: action being completed.
        """
        # Set apply_style to False because we don't want hints to look like errors
        super().__init__(_build_hint(parser, arg_action), apply_style=False)


class ArgparseCompleter:
    """Automatic command line completion based on argparse parameters."""

    def __init__(
        self,
        parser: Cmd2ArgumentParser,
        cmd_app: "Cmd",
        *,
        parent_tokens: Mapping[str, MutableSequence[str]] | None = None,
    ) -> None:
        """Create an ArgparseCompleter.

        :param parser: Cmd2ArgumentParser instance
        :param cmd_app: reference to the cmd2.Cmd instance that owns this ArgparseCompleter
        :param parent_tokens: optional Mapping of parent parsers' arg names to their tokens
                              This is only used by ArgparseCompleter when recursing on subcommand parsers
                              Defaults to None
        """
        self._parser = parser
        self._cmd_app = cmd_app

        if parent_tokens is None:
            parent_tokens = {}
        self._parent_tokens = parent_tokens

        # All flags in this command
        self._flags: list[str] = []

        # Maps flags to the argparse action object
        self._flag_to_action: dict[str, argparse.Action] = {}

        # Actions for positional arguments (by position index)
        self._positional_actions: list[argparse.Action] = []

        # This will be set if self._parser has subcommands
        self._subcommand_action: argparse._SubParsersAction[Cmd2ArgumentParser] | None = None

        # Start digging through the argparse structures.
        # _actions is the top level container of parameter definitions
        for action in self._parser._actions:
            # if the parameter is flag based, it will have option_strings
            if action.option_strings:
                # record each option flag
                for option in action.option_strings:
                    self._flags.append(option)
                    self._flag_to_action[option] = action

            # Otherwise this is a positional parameter
            else:
                self._positional_actions.append(action)
                # Check if this action defines subcommands
                if isinstance(action, argparse._SubParsersAction):
                    self._subcommand_action = action

    def complete(
        self,
        text: str,
        line: str,
        begidx: int,
        endidx: int,
        tokens: Sequence[str],
        *,
        cmd_set: CommandSet[Any] | None = None,
    ) -> Completions:
        """Complete text using argparse metadata.

        :param text: the string prefix we are attempting to match (all matches must begin with it)
        :param line: the current input line with leading whitespace removed
        :param begidx: the beginning index of the prefix text
        :param endidx: the ending index of the prefix text
        :param tokens: Sequence of argument tokens being passed to the parser
        :param cmd_set: if completing a command, the CommandSet the command's function belongs to, if applicable.
                        Defaults to None.
        :return: a Completions object
        :raises CompletionError: for various types of completion errors
        """
        if not tokens:
            return Completions()

        # Positionals args that are left to parse
        remaining_positionals = deque(self._positional_actions)

        # This gets set to True when flags will no longer be processed as argparse flags
        # That can happen when -- is used or an argument with nargs=argparse.REMAINDER is used
        skip_remaining_flags = False

        # _ArgumentState of the current positional
        pos_arg_state: _ArgumentState | None = None

        # _ArgumentState of the current flag
        flag_arg_state: _ArgumentState | None = None

        # Non-reusable flags that we've parsed
        used_flags: set[str] = set()

        # Keeps track of arguments we've seen and any tokens they consumed
        consumed_arg_values: dict[str, list[str]] = {}

        # Completed mutually exclusive groups
        completed_mutex_groups: dict[argparse._MutuallyExclusiveGroup, argparse.Action] = {}

        def consume_argument(arg_state: _ArgumentState, arg_token: str) -> None:
            """Consume token as an argument."""
            arg_state.count += 1
            consumed_arg_values.setdefault(arg_state.action.dest, []).append(arg_token)

        #############################################################################################
        # Parse all but the last token
        #############################################################################################
        for token_index, token in enumerate(tokens[:-1]):
            # If we're in a positional REMAINDER arg, force all future tokens to go to that
            if pos_arg_state is not None and pos_arg_state.is_remainder:
                consume_argument(pos_arg_state, token)
                continue

            # If we're in a flag REMAINDER arg, force all future tokens to go to that until a double dash is hit
            if flag_arg_state is not None and flag_arg_state.is_remainder:
                if token == "--":  # noqa: S105
                    flag_arg_state = None
                else:
                    consume_argument(flag_arg_state, token)
                continue

            # Handle '--' which tells argparse all remaining arguments are non-flags
            if token == "--" and not skip_remaining_flags:  # noqa: S105
                # Check if there is an unfinished flag
                if (
                    flag_arg_state is not None
                    and isinstance(flag_arg_state.min, int)
                    and flag_arg_state.count < flag_arg_state.min
                ):
                    raise _UnfinishedFlagError(flag_arg_state)

                # Otherwise end the current flag
                flag_arg_state = None
                skip_remaining_flags = True
                continue

            # Check if token is a flag
            if _looks_like_flag(token, self._parser) and not skip_remaining_flags:
                # Check if there is an unfinished flag
                if (
                    flag_arg_state is not None
                    and isinstance(flag_arg_state.min, int)
                    and flag_arg_state.count < flag_arg_state.min
                ):
                    raise _UnfinishedFlagError(flag_arg_state)

                # Reset flag arg state but not positional tracking because flags can be
                # interspersed anywhere between positionals
                flag_arg_state = None
                action = None

                # Does the token match a known flag?
                if token in self._flag_to_action:
                    action = self._flag_to_action[token]
                elif self._parser.allow_abbrev:
                    candidates_flags = [flag for flag in self._flag_to_action if flag.startswith(token)]
                    if len(candidates_flags) == 1:
                        action = self._flag_to_action[candidates_flags[0]]

                if action is not None:
                    self._update_mutex_groups(action, completed_mutex_groups, used_flags, remaining_positionals)

                    # Check if the action type allows the same flag to be provided multiple times.
                    # Reusable actions (append, count, extend) preserve their history so the
                    # completion logic knows which values have already been 'consumed'.
                    if not isinstance(
                        action,
                        (
                            argparse._AppendAction,
                            argparse._AppendConstAction,
                            argparse._CountAction,
                            argparse._ExtendAction,
                        ),
                    ):
                        # For standard 'overwrite' actions (e.g., --store), providing the flag
                        # again resets its state. We mark the flags as 'used' to potentially
                        # filter them from future completion results and clear any previously
                        # recorded values for this destination.
                        used_flags.update(action.option_strings)
                        consumed_arg_values[action.dest] = []

                    new_arg_state = _ArgumentState(action)

                    # Keep track of this flag if it can receive arguments
                    if new_arg_state.max > 0:
                        flag_arg_state = new_arg_state
                        skip_remaining_flags = flag_arg_state.is_remainder

            # Check if token is a flag's argument
            elif flag_arg_state is not None:
                consume_argument(flag_arg_state, token)

                # Check if we have finished with this flag
                if flag_arg_state.count >= flag_arg_state.max:
                    flag_arg_state = None

            # Otherwise treat token as a positional argument
            else:
                # If we aren't current tracking a positional, then get the next positional arg to handle this token
                if pos_arg_state is None and remaining_positionals:
                    action = remaining_positionals.popleft()

                    # Are we at a subcommand? If so, forward to the matching completer
                    if self._subcommand_action is not None and action == self._subcommand_action:
                        if token in self._subcommand_action.choices:
                            parent_tokens = {**self._parent_tokens, **consumed_arg_values}

                            # Include the subcommand name if its destination was set
                            if action.dest != argparse.SUPPRESS:
                                parent_tokens[action.dest] = [token]

                            parser = self._subcommand_action.choices[token]
                            completer = parser.completer_class(parser, self._cmd_app, parent_tokens=parent_tokens)
                            return completer.complete(text, line, begidx, endidx, tokens[token_index + 1 :], cmd_set=cmd_set)

                        # Invalid subcommand entered, so no way to complete remaining tokens
                        return Completions()

                    # Otherwise keep track of the argument
                    pos_arg_state = _ArgumentState(action)

                # Check if we have a positional to consume this token
                if pos_arg_state is not None:
                    self._update_mutex_groups(pos_arg_state.action, completed_mutex_groups, used_flags, remaining_positionals)
                    consume_argument(pos_arg_state, token)

                    # No more flags are allowed if this is a REMAINDER argument
                    if pos_arg_state.is_remainder:
                        skip_remaining_flags = True

                    # Check if we have finished with this positional
                    elif pos_arg_state.count >= pos_arg_state.max:
                        pos_arg_state = None

                        # Check if the next positional has nargs set to argparse.REMAINDER.
                        # At this point argparse allows no more flags to be processed.
                        if remaining_positionals and remaining_positionals[0].nargs == argparse.REMAINDER:
                            skip_remaining_flags = True

        #############################################################################################
        # We have parsed all but the last token and have enough information to complete it
        #############################################################################################
        return self._handle_last_token(
            text,
            line,
            begidx,
            endidx,
            flag_arg_state,
            pos_arg_state,
            remaining_positionals,
            consumed_arg_values,
            used_flags,
            skip_remaining_flags,
            cmd_set,
        )

    def _update_mutex_groups(
        self,
        arg_action: argparse.Action,
        completed_mutex_groups: dict[argparse._MutuallyExclusiveGroup, argparse.Action],
        used_flags: set[str],
        remaining_positionals: deque[argparse.Action],
    ) -> None:
        """Manage mutually exclusive group constraints and argument pruning for a given action.

        If an action belongs to a mutually exclusive group, this method ensures no other member
        has been used and updates the parser state to "consume" all remaining conflicting arguments.

        :raises CompletionError: if another member of the same mutually exclusive group
                                 has already been used.
        """
        # Check if this action is in a mutually exclusive group
        for group in self._parser._mutually_exclusive_groups:
            if arg_action in group._group_actions:
                # Check if the group this action belongs to has already been completed
                if group in completed_mutex_groups:
                    # If this is the action that completed the group, then there is no error
                    # since it's allowed to appear on the command line more than once.
                    completer_action = completed_mutex_groups[group]
                    if arg_action == completer_action:
                        return

                    arg_str = f"{argparse._get_action_name(arg_action)}"
                    completer_str = f"{argparse._get_action_name(completer_action)}"
                    error = f"Error: argument {arg_str}: not allowed with argument {completer_str}"
                    raise CompletionError(error)

                # Mark that this action completed the group
                completed_mutex_groups[group] = arg_action

                # Don't complete any of the other args in the group
                for group_action in group._group_actions:
                    if group_action == arg_action:
                        continue
                    if group_action in self._flag_to_action.values():
                        used_flags.update(group_action.option_strings)
                    elif group_action in remaining_positionals:
                        remaining_positionals.remove(group_action)

                # Arg can only be in one group, so we are done
                break

    def _handle_last_token(
        self,
        text: str,
        line: str,
        begidx: int,
        endidx: int,
        flag_arg_state: _ArgumentState | None,
        pos_arg_state: _ArgumentState | None,
        remaining_positionals: deque[argparse.Action],
        consumed_arg_values: dict[str, list[str]],
        used_flags: set[str],
        skip_remaining_flags: bool,
        cmd_set: CommandSet[Any] | None,
    ) -> Completions:
        """Perform final completion step handling positionals and flags."""
        # Check if we are completing a flag name. This check ignores strings with a length of one, like '-'.
        # This is because that could be the start of a negative number which may be a valid completion for
        # the current argument. We will handle the completion of flags that start with only one prefix
        # character (-f) at the end.
        if _looks_like_flag(text, self._parser) and not skip_remaining_flags:
            if (
                flag_arg_state is not None
                and isinstance(flag_arg_state.min, int)
                and flag_arg_state.count < flag_arg_state.min
            ):
                raise _UnfinishedFlagError(flag_arg_state)
            return self._complete_flags(text, line, begidx, endidx, used_flags)

        # Check if we are completing a flag's argument
        if flag_arg_state is not None:
            completions = self._complete_arg(text, line, begidx, endidx, flag_arg_state, consumed_arg_values, cmd_set=cmd_set)

            # If we have results, then return them
            if completions:
                return completions

            # Otherwise, print a hint if the flag isn't finished or text isn't possibly the start of a flag
            if (
                (isinstance(flag_arg_state.min, int) and flag_arg_state.count < flag_arg_state.min)
                or not _single_prefix_char(text, self._parser)
                or skip_remaining_flags
            ):
                raise _NoResultsError(self._parser, flag_arg_state.action)

        # Otherwise check if we have a positional to complete
        elif pos_arg_state is not None or remaining_positionals:
            # If we aren't current tracking a positional, then get the next positional arg to handle this token
            if pos_arg_state is None:
                action = remaining_positionals.popleft()
                pos_arg_state = _ArgumentState(action)

            completions = self._complete_arg(text, line, begidx, endidx, pos_arg_state, consumed_arg_values, cmd_set=cmd_set)

            # If we have results, then return them
            if completions:
                return completions

            # Otherwise, print a hint if text isn't possibly the start of a flag
            if not _single_prefix_char(text, self._parser) or skip_remaining_flags:
                raise _NoResultsError(self._parser, pos_arg_state.action)

        # If we aren't skipping remaining flags, then complete flag names if either is True:
        #   1. text is a single flag prefix character that didn't complete against any argument values
        #   2. there are no more positionals to complete
        if not skip_remaining_flags and (_single_prefix_char(text, self._parser) or not remaining_positionals):
            return self._complete_flags(text, line, begidx, endidx, used_flags)

        return Completions()

    def _complete_flags(self, text: str, line: str, begidx: int, endidx: int, used_flags: set[str]) -> Completions:
        """Completion routine for a parsers unused flags."""
        # Build a list of flags that can be completed
        match_against: list[str] = []

        for flag in self._flags:
            # Make sure this flag hasn't already been used
            if flag not in used_flags:
                # Make sure this flag isn't considered hidden
                action = self._flag_to_action[flag]
                if action.help != argparse.SUPPRESS:
                    match_against.append(flag)

        # Build a dictionary linking actions with their matched flag names
        matched_actions: dict[argparse.Action, list[str]] = {}

        # Keep flags sorted in the order provided by argparse so our completion
        # suggestions display the same as argparse help text.
        matched_flags = self._cmd_app.basic_complete(text, line, begidx, endidx, match_against, sort=False)

        for flag in matched_flags.to_strings():
            action = self._flag_to_action[flag]
            matched_actions.setdefault(action, []).append(flag)

        # For completion suggestions, group matched flags by action
        items: list[CompletionItem] = []
        for action, option_strings in matched_actions.items():
            flag_text = ", ".join(option_strings)

            # Mark optional flags with brackets
            if not action.required:
                flag_text = "[" + flag_text + "]"

            # Use the first option string as the completion result for this action
            items.append(
                CompletionItem(
                    option_strings[0],
                    display=flag_text,
                    display_meta=action.help or "",
                )
            )

        return Completions(items)

    @staticmethod
    def _validate_table_data(arg_state: _ArgumentState, completions: Completions) -> None:
        """Verify the integrity of completion table data.

        :raises ValueError: if there is an error with the data.
        """
        table_columns = arg_state.action.get_table_columns()  # type: ignore[attr-defined]
        has_table_data = any(item.table_data for item in completions)

        if table_columns is None:
            if has_table_data:
                raise ValueError(
                    f"Argument '{arg_state.action.dest}' has CompletionItems with table_data, "
                    f"but no table_columns were defined in add_argument()."
                )
            return

        # If columns are defined, then every item must have data, and lengths must match
        for item in completions:
            if not item.table_data:
                raise ValueError(
                    f"Argument '{arg_state.action.dest}' has table_columns defined, "
                    f"but the CompletionItem for '{item.text}' is missing table_data."
                )
            if len(item.table_data) != len(table_columns):
                raise ValueError(
                    f"Argument '{arg_state.action.dest}': table_data length ({len(item.table_data)}) "
                    f"does not match table_columns length ({len(table_columns)}) for item '{item.text}'."
                )

    def _build_completion_table(self, arg_state: _ArgumentState, completions: Completions) -> Completions:
        """Build a rich.Table for completion results if applicable."""
        # Verify integrity of completion data
        self._validate_table_data(arg_state, completions)

        table_columns = cast(
            Sequence[str | Column] | None,
            arg_state.action.get_table_columns(),  # type: ignore[attr-defined]
        )

        # Skip table generation if results are outside thresholds or no columns are defined
        if (
            len(completions) < 2
            or len(completions) > self._cmd_app.max_completion_table_items
            or table_columns is None
        ):  # fmt: skip
            return completions

        # If a metavar was defined, use that instead of the dest field
        destination = arg_state.action.metavar or arg_state.action.dest

        # Handle case where metavar was a tuple
        if isinstance(destination, tuple):
            # Figure out what string in the tuple to use based on how many of the arguments have been completed.
            # Use min() to avoid going passed the end of the tuple to support nargs being ZERO_OR_MORE and
            # ONE_OR_MORE. In those cases, argparse limits metavar tuple to 2 elements but we may be completing
            # the 3rd or more argument here.
            destination = destination[min(len(destination) - 1, arg_state.count)]

        # Build header row
        rich_columns: list[Column] = []
        rich_columns.append(
            Column(
                destination.upper(),
                justify="right" if completions.numeric_display else "left",
                no_wrap=True,
            )
        )
        rich_columns.extend(
            column if isinstance(column, Column) else Column(column, overflow="fold") for column in table_columns
        )

        # Build the table
        table = Cmd2SimpleTable(*rich_columns)
        for item in completions:
            table.add_row(Text.from_ansi(item.display), *item.table_data)

        return dataclasses.replace(
            completions,
            table=table,
        )

    def complete_subcommand_help(self, text: str, line: str, begidx: int, endidx: int, tokens: Sequence[str]) -> Completions:
        """Supports cmd2's help command in the completion of subcommand names.

        :param text: the string prefix we are attempting to match (all matches must begin with it)
        :param line: the current input line with leading whitespace removed
        :param begidx: the beginning index of the prefix text
        :param endidx: the ending index of the prefix text
        :param tokens: arguments passed to command/subcommand
        :return: a Completions object
        """
        # If our parser has subcommands, we must examine the tokens and check if they are subcommands
        # If so, we will let the subcommand's parser handle the rest of the tokens via another ArgparseCompleter.
        if self._subcommand_action is not None:
            for token_index, token in enumerate(tokens):
                if token in self._subcommand_action.choices:
                    parser = self._subcommand_action.choices[token]
                    completer = parser.completer_class(parser, self._cmd_app)
                    return completer.complete_subcommand_help(text, line, begidx, endidx, tokens[token_index + 1 :])

                if token_index == len(tokens) - 1:
                    # Since this is the last token, we will attempt to complete it
                    return self._cmd_app.basic_complete(text, line, begidx, endidx, self._subcommand_action.choices)
                break
        return Completions()

    def print_help(self, tokens: Sequence[str], file: IO[str] | None = None) -> None:
        """Supports cmd2's help command in the printing of help text.

        :param tokens: arguments passed to help command
        :param file: optional file object where the argparse should write help text
                     If not supplied, argparse will write to sys.stdout.
        """
        # If our parser has subcommands, we must examine the tokens and check if they are subcommands.
        # If so, we will let the subcommand's parser handle the rest of the tokens via another ArgparseCompleter.
        if tokens and self._subcommand_action is not None:
            parser = self._subcommand_action.choices.get(tokens[0])
            if parser is not None:
                completer = parser.completer_class(parser, self._cmd_app)
                completer.print_help(tokens[1:], file)
                return
        self._parser.print_help(file)

    def _choices_to_items(self, arg_state: _ArgumentState) -> list[CompletionItem]:
        """Convert an action's choices to a list of CompletionItems."""
        if arg_state.action.choices is None:
            return []

        # If choices are subcommands, then get their help text to populate display_meta.
        if isinstance(arg_state.action, argparse._SubParsersAction):
            parser_help = {}
            for action in arg_state.action._choices_actions:
                if action.dest in arg_state.action.choices:
                    subparser = arg_state.action.choices[action.dest]
                    parser_help[subparser] = action.help or ""

            return [
                CompletionItem(name, display_meta=parser_help.get(subparser, ""))
                for name, subparser in arg_state.action.choices.items()
            ]

        # Standard choices
        return [
            choice if isinstance(choice, CompletionItem) else CompletionItem(choice) for choice in arg_state.action.choices
        ]

    def _prepare_callable_params(
        self,
        to_call: UnboundChoicesProvider[CmdOrSetT] | UnboundCompleter[CmdOrSetT],
        arg_state: _ArgumentState,
        text: str,
        consumed_arg_values: dict[str, list[str]],
        cmd_set: CommandSet[Any] | None,
    ) -> tuple[list[Any], dict[str, Any]]:
        """Resolve the instance and arguments required to call a choices/completer function."""
        args: list[Any] = []
        kwargs: dict[str, Any] = {}

        # Resolve the 'self' instance for the method
        self_arg = self._cmd_app._resolve_func_self(to_call, cmd_set)
        if self_arg is None:
            raise CompletionError("Could not find CommandSet instance matching defining type")

        args.append(self_arg)

        # Check if the function expects 'arg_tokens'
        to_call_params = inspect.signature(to_call).parameters
        if ARG_TOKENS in to_call_params:
            arg_tokens = {**self._parent_tokens, **consumed_arg_values}
            arg_tokens.setdefault(arg_state.action.dest, []).append(text)
            kwargs[ARG_TOKENS] = arg_tokens

        return args, kwargs

    def _complete_arg(
        self,
        text: str,
        line: str,
        begidx: int,
        endidx: int,
        arg_state: _ArgumentState,
        consumed_arg_values: dict[str, list[str]],
        *,
        cmd_set: CommandSet[Any] | None = None,
    ) -> Completions:
        """Completion routine for an argparse argument.

        :return: a Completions object
        :raises CompletionError: if the completer or choices function this calls raises one
        """
        # Check if the argument uses a completer
        completer = arg_state.action.get_completer()  # type: ignore[attr-defined]
        if completer is not None:
            args, kwargs = self._prepare_callable_params(
                completer,
                arg_state,
                text,
                consumed_arg_values,
                cmd_set,
            )
            args.extend([text, line, begidx, endidx])
            completions: Completions = completer(*args, **kwargs)

        # Otherwise it uses a choices provider or choices list
        else:
            choices_provider = arg_state.action.get_choices_provider()  # type: ignore[attr-defined]
            if choices_provider is not None:
                args, kwargs = self._prepare_callable_params(
                    choices_provider,
                    arg_state,
                    text,
                    consumed_arg_values,
                    cmd_set,
                )
                all_choices = list(choices_provider(*args, **kwargs))
                sort = False  # choices_provider results are already sorted, so don't re-sort them
            else:
                all_choices = self._choices_to_items(arg_state)
                sort = True

            # Filter used values and run basic completion
            used_values = consumed_arg_values.get(arg_state.action.dest, [])
            filtered = [choice for choice in all_choices if choice.text not in used_values]
            completions = self._cmd_app.basic_complete(text, line, begidx, endidx, filtered, sort=sort)

        return self._build_completion_table(arg_state, completions)


# The default ArgparseCompleter class for a cmd2 app
DEFAULT_ARGPARSE_COMPLETER: type[ArgparseCompleter] = ArgparseCompleter


def set_default_argparse_completer(completer_class: type[ArgparseCompleter]) -> None:
    """Set the default ArgparseCompleter class for a cmd2 app.

    :param completer_class: Type that is a subclass of ArgparseCompleter.
    """
    global DEFAULT_ARGPARSE_COMPLETER  # noqa: PLW0603
    DEFAULT_ARGPARSE_COMPLETER = completer_class



"""Bridges calls made inside of a Python environment to the Cmd2 host app.

Maintains a reasonable degree of isolation between the two.
"""

import sys
from contextlib import redirect_stderr
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    NamedTuple,
    TextIO,
    cast,
)

from .utils import StdSim  # namedtuple_with_defaults,

if TYPE_CHECKING:  # pragma: no cover
    from .cmd2 import Cmd


class CommandResult(NamedTuple):
    """Encapsulates the results from a cmd2 app command.

    :stdout: str - output captured from stdout while this command is executing
    :stderr: str - output captured from stderr while this command is executing
    :stop: bool - return value of onecmd_plus_hooks after it runs the given
           command line.
    :data: possible data populated by the command.

    Any combination of these fields can be used when developing a scripting API
    for a given command. By default stdout, stderr, and stop will be captured
    for you. If there is additional command specific data, then write that to
    cmd2's last_result member. That becomes the data member of this tuple.

    In some cases, the data member may contain everything needed for a command
    and storing stdout and stderr might just be a duplication of data that
    wastes memory. In that case, the StdSim can be told not to store output
    with its pause_storage member. While this member is True, any output sent
    to StdSim won't be saved in its buffer.

    The code would look like this::

        if isinstance(self.stdout, StdSim):
            self.stdout.pause_storage = True

        if isinstance(sys.stderr, StdSim):
            sys.stderr.pause_storage = True

    See [cmd2.utils.StdSim][] for more information.

    .. note::

       Named tuples are immutable. The contents are there for access,
       not for modification.
    """

    stdout: str = ""
    stderr: str = ""
    stop: bool = False
    data: Any = None

    def __bool__(self) -> bool:
        """Return True if the command succeeded, otherwise False."""
        # If data was set, then use it to determine success
        if self.data is not None:
            return bool(self.data)

        # Otherwise check if stderr was filled out
        return not self.stderr


class PyBridge:
    """Provides a Python API wrapper for application commands.

    :param cmd_app: app being controlled by this PyBridge.
    :param add_to_history: If True, then add all commands run by this PyBridge to history.
                           Defaults to True.
    """

    def __init__(self, cmd_app: "Cmd", *, add_to_history: bool = True) -> None:
        """Initialize PyBridge instances."""
        self._cmd_app = cmd_app
        self._add_to_history = add_to_history
        self.cmd_echo = False

        # Tells if any of the commands run via __call__ returned True for stop
        self.stop = False

    def __dir__(self) -> list[str]:
        """Return a custom set of attribute names."""
        attributes: list[str] = []
        attributes.insert(0, "cmd_echo")
        return attributes

    def __call__(self, command: str, *, echo: bool | None = None) -> CommandResult:
        """Provide functionality to call application commands by calling PyBridge.

        ex: app('help')
        :param command: command line being run
        :param echo: If provided, this temporarily overrides the value of self.cmd_echo
                     while the command runs. If True, output will be echoed to _cmd_app.stdout
                     and sys.stderr. (Defaults to None)
        """
        if echo is None:
            echo = self.cmd_echo

        # This will be used to capture _cmd_app.stdout
        copy_cmd_stdout = StdSim(cast(TextIO | StdSim, self._cmd_app.stdout), echo=echo)

        # Pause the storing of stdout until onecmd_plus_hooks enables it
        copy_cmd_stdout.pause_storage = True

        # This will be used to capture sys.stderr
        copy_stderr = StdSim(sys.stderr, echo=echo)

        self._cmd_app.last_result = None

        stop = False
        try:
            self._cmd_app.stdout = cast(TextIO, copy_cmd_stdout)

            with redirect_stderr(cast(IO[str], copy_stderr)):
                stop = self._cmd_app.onecmd_plus_hooks(
                    command,
                    add_to_history=self._add_to_history,
                    py_bridge_call=True,
                )
        finally:
            with self._cmd_app.sigint_protection:
                self._cmd_app.stdout = cast(TextIO, copy_cmd_stdout.inner_stream)
                self.stop = stop or self.stop

        # Save the result
        return CommandResult(
            stdout=copy_cmd_stdout.getvalue(),
            stderr=copy_stderr.getvalue(),
            stop=stop,
            data=self._cmd_app.last_result,
        )
